from typing import List, Optional, Dict, Any, Union, Tuple
from uuid import uuid4
import json

import pandas as pd
import dash
from dash.dependencies import Input, Output
from dash.exceptions import PreventUpdate
import dash_html_components as html
import dash_core_components as dcc
import webviz_core_components as wcc
from webviz_config.common_cache import CACHE

from .._abbreviations.number_formatting import si_prefixed


class TornadoPlot:
    """### TornadoPlot

    This private plugin visualizes a Tornado plot.
    It is meant to be used as a component in other plugin, and is initialized
     with a dataframe of realizations with corresponding sensitivities,
    but without the response values that are to be plotted.
    Instead we registers a dcc.Store which will contain the response values.

    To use:
    1. Initialize an instance of this class in a plugin.
    2. Add tornadoplot.layout to the plugin layout
    3. Register a callback that writes a json dump to tornadoplot.storage_id
    The format of the json dump must be ('ENSEMBLE' and 'data' are mandatory, the others optional):
    {'ENSEMBLE': name of ensemble,
     'data': 2d array of realizations / response values
     'number_format' (str): Format of the numeric part based on the Python Format Specification
      Mini-Language e.g. '#.3g' for 3 significant digits, '.2f' for two decimals, or '.0f' for no
      decimals.
     'unit' (str): String to append at the end as a unit.
     'spaced' (bool): Include a space between last numerical digit and SI-prefix.
     'locked_si_prefix' (str or int): Lock the SI prefix to either a string (e.g. 'm' (milli) or 'M'
      (mega)), or an integer which is the base 10 exponent (e.g. 3 for kilo, -3 for milli).
    }

    Mouse events:
    The current case at mouse cursor can be retrieved by registering a callback
    that reads from  `tornadoplot.click_id` if `allow_click` has been specified at initialization.


    * `realizations`: Dataframe of realizations with corresponding sensitivity cases
    * `reference`: Which sensitivity to use as reference.
    * `allow_click`: Registers a callback to store current data on mouse click
    """

    def __init__(
        self,
        app: dash.Dash,
        realizations: pd.DataFrame,
        reference: str = "rms_seed",
        allow_click: bool = False,
    ):

        self.realizations = realizations
        self.sensnames = list(self.realizations["SENSNAME"].unique())
        if self.sensnames == [None]:
            raise KeyError(
                "No sensitivity information found in ensemble. "
                "Containers utilizing tornadoplot can only be used for ensembles with "
                "one by one design matrix setup "
                "(SENSNAME and SENSCASE must be present in parameter file)."
            )
        self.initial_reference = (
            reference if reference in self.sensnames else self.sensnames[0]
        )
        self.allow_click = allow_click
        self.uid = uuid4()
        self.plotly_theme = app.webviz_settings["theme"].plotly_theme
        self.set_callbacks(app)

    def ids(self, element: str) -> str:
        """Generate unique id for dom element"""
        return f"{element}-id-{self.uid}"

    @property
    def tour_steps(self) -> List[Dict[str, Any]]:
        return [
            {
                "id": self.ids("tornado-graph"),
                "content": ("Shows tornado plot."),
            },
            {
                "id": self.ids("reference"),
                "content": (
                    "Set reference sensitivity for which to calculate tornado plot"
                ),
            },
            {
                "id": self.ids("scale"),
                "content": (
                    "Set tornadoplot scale to either percentage or absolute values"
                ),
            },
            {
                "id": self.ids("cut-by-ref"),
                "content": (
                    "Remove sensitivities smaller than the reference from the plot"
                ),
            },
            {
                "id": self.ids("reset"),
                "content": "Clears the currently selected sensitivity",
            },
        ]

    @property
    def storage_id(self) -> str:
        """The id of the dcc.Store component that holds the tornado data"""
        return self.ids("storage")

    @property
    def click_id(self) -> str:
        """The id of the dcc.Store component that holds click data"""
        return self.ids("click-store")

    @property
    def high_low_storage_id(self) -> str:
        """The id of the dcc.Store component that holds click data"""
        return self.ids("high-low-storage")

    @staticmethod
    def set_grid_layout(columns: Union[str, List[str]]) -> Dict[str, str]:
        return {
            "display": "grid",
            "alignContent": "space-around",
            "justifyContent": "space-between",
            "gridTemplateColumns": f"{columns}",
        }

    @property
    def layout(self) -> html.Div:
        return html.Div(
            [
                html.Div(
                    style={"marginLeft": "10px"},
                    children=[
                        html.Label(
                            "Tornado Plot",
                            style={"textAlign": "center", "font-weight": "bold"},
                        ),
                        html.Div(
                            style=self.set_grid_layout("1fr 1fr"),
                            children=[
                                html.Label("Reference:"),
                                html.Label("Scale:"),
                            ],
                        ),
                        html.Div(
                            style=self.set_grid_layout("1fr 1fr"),
                            children=[
                                dcc.Dropdown(
                                    id=self.ids("reference"),
                                    options=[
                                        {"label": r, "value": r} for r in self.sensnames
                                    ],
                                    value=self.initial_reference,
                                    clearable=False,
                                    persistence=True,
                                    persistence_type="session",
                                ),
                                dcc.Dropdown(
                                    id=self.ids("scale"),
                                    options=[
                                        {"label": r, "value": r}
                                        for r in ["Percentage", "Absolute"]
                                    ],
                                    value="Percentage",
                                    clearable=False,
                                    persistence=True,
                                    persistence_type="session",
                                ),
                            ],
                        ),
                        dcc.Checklist(
                            id=self.ids("cut-by-ref"),
                            options=[
                                {
                                    "label": "Cut by reference",
                                    "value": "Cut by reference",
                                },
                            ],
                            value=[],
                            persistence=True,
                            persistence_type="session",
                        ),
                        html.Details(
                            open=False,
                            children=[
                                html.Summary("Filter"),
                                wcc.Select(
                                    id=self.ids("sens_filter"),
                                    options=[
                                        {"label": i, "value": i} for i in self.sensnames
                                    ],
                                    value=self.sensnames,
                                    multi=True,
                                    size=min(10, len(self.sensnames)),
                                    persistence=True,
                                    persistence_type="session",
                                ),
                            ],
                        ),
                        html.Button(
                            style={
                                "position": "relative",
                                "top": "-50%",
                                "fontSize": "10px",
                            },
                            id=self.ids("reset"),
                            children="Clear selected",
                        ),
                        wcc.Graph(
                            id=self.ids("tornado-graph"),
                            config={"displayModeBar": False},
                        ),
                        dcc.Store(id=self.ids("storage"), storage_type="session"),
                        dcc.Store(id=self.ids("click-store"), storage_type="session"),
                        dcc.Store(
                            id=self.ids("high-low-storage"), storage_type="session"
                        ),
                    ],
                )
            ]
        )

    def set_callbacks(self, app: dash.Dash) -> None:
        @app.callback(
            [
                Output(self.ids("tornado-graph"), "figure"),
                Output(self.ids("high-low-storage"), "data"),
            ],
            [
                Input(self.ids("reference"), "value"),
                Input(self.ids("scale"), "value"),
                Input(self.ids("cut-by-ref"), "value"),
                Input(self.ids("storage"), "data"),
                Input(self.ids("sens_filter"), "value"),
            ],
        )
        def _calc_tornado(
            reference: str,
            scale: str,
            cutbyref: str,
            data: Union[str, bytes, bytearray],
            sens_filter: List[str],
        ) -> Tuple[dict, dict]:
            if not data:
                raise PreventUpdate
            data = json.loads(data)
            if not isinstance(data, dict):
                raise PreventUpdate
            values = pd.DataFrame(data["data"], columns=["REAL", "VALUE"])
            realizations = self.realizations.loc[
                self.realizations["ENSEMBLE"] == data["ENSEMBLE"]
            ]
            try:
                return tornado_plot(
                    realizations,
                    values,
                    sens_filter=sens_filter,
                    plotly_theme=self.plotly_theme,
                    reference=reference,
                    scale=scale,
                    cutbyref="Cut by reference" in cutbyref,
                    number_format=data.get("number_format", ""),
                    unit=data.get("unit", ""),
                    spaced=data.get("spaced", True),
                    locked_si_prefix=data.get("locked_si_prefix", None),
                )
            except KeyError:
                return {}, {}

        if self.allow_click:

            @app.callback(
                Output(self.ids("click-store"), "data"),
                [
                    Input(self.ids("tornado-graph"), "clickData"),
                    Input(self.ids("reset"), "n_clicks"),
                ],
            )
            def _save_click_data(data: dict, nclicks: Optional[int]) -> str:
                if dash.callback_context.triggered is None:
                    raise PreventUpdate

                ctx = dash.callback_context.triggered[0]["prop_id"].split(".")[0]

                if ctx == self.ids("reset") and nclicks:

                    return json.dumps(
                        {
                            "real_low": [],
                            "real_high": [],
                            "sens_name": None,
                        }
                    )
                try:
                    real_low = data["points"][0]["customdata"]
                    real_high = data["points"][1]["customdata"]
                    sens_name = data["points"][0]["y"]
                    return json.dumps(
                        {
                            "real_low": real_low,
                            "real_high": real_high,
                            "sens_name": sens_name,
                        }
                    )
                except TypeError as exc:
                    raise PreventUpdate from exc


def scale_to_ref(value: float, ref: float, scale: str) -> float:
    value_ref = value - ref
    if scale == "Percentage":
        value_ref = (100 * (value_ref / ref)) if ref != 0 else 0
    return value_ref


def sort_by_max(tornadotable: pd.DataFrame) -> pd.DataFrame:
    # TODO(Ruben): This mypy error seems to be a false positive?
    """ Sorts table based on max(abs('low', 'high')) """
    tornadotable["max"] = (
        tornadotable[["low", "high"]]
        .apply(lambda x: max(x.min(), x.max(), key=abs), axis=1)  # type: ignore
        .abs()
    )
    df_sorted = tornadotable.sort_values("max", ascending=True)
    df_sorted.drop(["max"], axis=1, inplace=True)
    return df_sorted


def cut_by_ref(tornadotable: pd.DataFrame, refname: str) -> pd.DataFrame:
    """ Removes sensitivities smaller than reference sensitivity from table """
    maskref = tornadotable.sensname == refname
    reflow = tornadotable[maskref].low.abs()
    refhigh = tornadotable[maskref].high.abs()
    refmax = max(float(reflow), float(refhigh))
    dfr_filtered = tornadotable.loc[
        (tornadotable["sensname"] == refname)
        | (
            (tornadotable["low"].abs() >= refmax)
            | (tornadotable["high"].abs() >= refmax)
        )
    ]
    return dfr_filtered


@CACHE.memoize(timeout=CACHE.TIMEOUT)
# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
def tornado_plot(
    realizations: pd.DataFrame,
    data: pd.DataFrame,
    sens_filter: List[str],
    plotly_theme: Dict[str, Any],
    reference: str = "rms_seed",
    scale: str = "Percentage",
    cutbyref: bool = True,
    number_format: str = "",
    unit: str = "",
    spaced: bool = True,
    locked_si_prefix: Optional[int] = None,
) -> Tuple[dict, dict]:

    # Raise key error if no senscases, i.e. the ensemble has no design matrix
    if list(realizations["SENSCASE"].unique()) == [None]:
        raise KeyError

    # Calculate average response value for reference sensitivity
    ref_avg = data.loc[
        data["REAL"].isin(
            realizations.loc[realizations["SENSNAME"] == reference]["REAL"]
        )
    ]["VALUE"].mean()

    df = calc_tornado_df(data, realizations, ref_avg, scale)

    locked_si_prefix_relative: Optional[int]

    # Drops sensitivities smaller than reference if specified
    if cutbyref and df["sensname"].str.contains(reference).any():
        df = cut_by_ref(df, reference)

    df = df.loc[df["sensname"].isin(sens_filter)]
    df = sort_by_max(df)

    store_low_high = {
        sensname: {
            "real_low": sens_name_df["low_reals"].tolist()[0],
            "real_high": sens_name_df["high_reals"].tolist()[0],
        }
        for sensname, sens_name_df in df.groupby(["sensname"])
    }

    # If percentage, unit is % and we turn off SI-prefix
    if scale == "Percentage":
        unit_x = "%"
        locked_si_prefix_relative = 0
    else:
        unit_x = unit
        locked_si_prefix_relative = locked_si_prefix
    # Return tornado data as Plotly figure
    plot_data = [
        dict(
            type="bar",
            y=df["sensname"],
            x=df["low"],
            name="low",
            base=df["low_base"],
            customdata=df["low_reals"],
            hovertext=[
                f"{si_prefixed(x, number_format, unit_x, spaced, locked_si_prefix_relative)}"
                f"<br>Case: {label}<br>True Value: "
                f"{si_prefixed(val, number_format, unit, spaced, locked_si_prefix)}"
                f"<br>Realizations: "
                f"{printable_int_list(reals)}"
                if reals
                else None
                for x, label, val, reals in zip(
                    df["low_tooltip"],
                    df["low_label"],
                    df["true_low"],
                    df["low_reals"],
                )
            ],
            hoverinfo="text",
            orientation="h",
        ),
        dict(
            type="bar",
            y=df["sensname"],
            x=df["high"],
            name="high",
            base=df["high_base"],
            customdata=df["high_reals"],
            hovertext=[
                f"{si_prefixed(x, number_format, unit_x, spaced, locked_si_prefix_relative)}"
                f"<br>Case: {label}<br>True Value: "
                f"{si_prefixed(val, number_format, unit, spaced, locked_si_prefix)}"
                f"<br>Realizations: "
                f"{printable_int_list(reals)}"
                if reals
                else None
                for x, label, val, reals in zip(
                    df["high_tooltip"],
                    df["high_label"],
                    df["true_high"],
                    df["high_reals"],
                )
            ],
            hoverinfo="text",
            orientation="h",
        ),
    ]
    layout: Dict[str, Any] = {}
    layout.update(plotly_theme["layout"])
    layout.update(
        {
            "barmode": "relative",
            "margin": {"l": 0, "r": 0, "b": 20, "t": 50},
            "xaxis": {
                "title": scale,
                "autorange": True,
                "showgrid": False,
                "zeroline": False,
                "showline": True,
                "automargin": True,
            },
            "yaxis": {
                "autorange": True,
                "showgrid": False,
                "zeroline": False,
                "showline": False,
                "automargin": True,
                "title": None,
                "dtick": 1,
            },
            "showlegend": False,
            "hovermode": "y",
            "annotations": [
                {
                    "x": 0,
                    "y": len(list(df["low"])),
                    "xref": "x",
                    "yref": "y",
                    "text": f"Reference avg: "
                    f"{si_prefixed(ref_avg, number_format, unit, spaced, locked_si_prefix)}",
                    "showarrow": True,
                    "align": "center",
                    "arrowhead": 2,
                    "arrowsize": 1,
                    "arrowwidth": 1,
                    "arrowcolor": "#636363",
                    "ax": 20,
                    "ay": -25,
                }
            ],
        }
    )
    return (
        {"data": plot_data, "layout": layout},
        store_low_high,
    )


# pylint: disable=too-many-locals
def calc_tornado_df(
    data: pd.DataFrame, realizations: pd.DataFrame, ref_avg: float, scale: str
) -> pd.DataFrame:
    # Group by sensitivity name/case and calculate average values for each case
    arr = []

    for sens_name, sens_name_df in realizations.groupby(["SENSNAME"]):
        # Excluding the reference case as well as any cases named `ref`
        # `ref` is used as `SENSNAME`, typically for a single realization only,
        # when no seed uncertainty is used
        if sens_name == "ref":
            continue

        # If `SENSTYPE` is scalar grab the mean for each `SENSCASE`
        if sens_name_df["SENSTYPE"].all() == "scalar":
            for sens_case, sens_case_df in sens_name_df.groupby(["SENSCASE"]):
                values = data.loc[data["REAL"].isin(sens_case_df["REAL"])][
                    "VALUE"
                ].mean()

                arr.append(
                    {
                        "sensname": sens_name,
                        "senscase": sens_case,
                        "values": values,
                        "values_ref": scale_to_ref(values, ref_avg, scale),
                        "reals": list(map(int, sens_case_df["REAL"])),
                    }
                )
        # If `SENSTYPE` is monte carlo get p10, p90
        elif sens_name_df["SENSTYPE"].all() == "mc":
            # Get data for relevant realizations
            case_df = data.loc[data["REAL"].isin(sens_name_df["REAL"])]

            # Calculate p90(low) and p10(high)
            p90 = case_df["VALUE"].quantile(0.10)
            p10 = case_df["VALUE"].quantile(0.90)

            # Extract list of realizations with values less then reference avg (low)
            low_reals = list(map(int, case_df.loc[case_df["VALUE"] <= ref_avg]["REAL"]))

            # Extract list of realizations with values higher then reference avg (high)
            high_reals = list(map(int, case_df.loc[case_df["VALUE"] > ref_avg]["REAL"]))

            arr.append(
                {
                    "sensname": sens_name,
                    "senscase": "P90",
                    "values": p90,
                    "values_ref": scale_to_ref(p90, ref_avg, scale),
                    "reals": low_reals,
                }
            )
            arr.append(
                {
                    "sensname": sens_name,
                    "senscase": "P10",
                    "values": p10,
                    "values_ref": scale_to_ref(p10, ref_avg, scale),
                    "reals": high_reals,
                }
            )
        else:
            raise ValueError(
                f"Sensitivities should be either 'scalar'or 'mc'. \
                Sensitivity: '{sens_name}' is neither."
            )

    # Group by sensitivity name and calculate low / high values
    arr2 = []
    for sensname, sens_name_df in pd.DataFrame(arr).groupby(["sensname"]):
        low = sens_name_df.copy().loc[sens_name_df["values_ref"].idxmin()]
        high = sens_name_df.copy().loc[sens_name_df["values_ref"].idxmax()]
        if sens_name_df["senscase"].nunique() == 1:
            # Single case sens, implies low == high, but testing just in case:
            if low["values_ref"] != high["values_ref"]:
                raise ValueError(
                    "For a single sensitivity case, low and high cases should be equal. Likely bug"
                )
            if low["values_ref"] < 0:
                # To avoid warnings for changing values of dataframe slices.
                high = high.copy()
                high["values_ref"] = 0
                high["reals"] = []
                high["senscase"] = None
                high["values"] = ref_avg
            else:
                low = (
                    low.copy()
                )  # To avoid warnings for changing values of dataframe slices.
                low["values_ref"] = 0
                low["reals"] = []
                low["senscase"] = None
                low["values"] = ref_avg

        arr2.append(
            {
                "low": calc_low_x(low["values_ref"], high["values_ref"]),
                "low_base": calc_low_base(low["values_ref"], high["values_ref"]),
                "low_label": low["senscase"],
                "low_tooltip": low["values_ref"],
                "true_low": low["values"],
                "low_reals": low["reals"],
                "sensname": sensname,
                "high": calc_high_x(low["values_ref"], high["values_ref"]),
                "high_base": calc_high_base(low["values_ref"], high["values_ref"]),
                "high_label": high["senscase"],
                "high_tooltip": high["values_ref"],
                "true_high": high["values"],
                "high_reals": high["reals"],
            }
        )
    return pd.DataFrame(arr2)


def calc_low_base(low: float, high: float) -> float:
    """
    From the low and high value of a parameter,
    calculates the base (starting x value) of the
    bar visualizing low values.
    >>> calc_low_base(1, 2)
    1
    >>> calc_low_base(-2, -1)
    -1
    >>> calc_low_base(-1, 1)
    0
    """
    if low < 0:
        return min(0, high)
    return low


def calc_high_base(low: float, high: float) -> float:
    """
    From the low and high value of a parameter,
    calculates the base (starting x value) of the bar
    visualizing high values.
    >>> calc_high_base(1, 2)
    1
    >>> calc_high_base(-1, 1)
    0
    >>> calc_high_base(1, 2)
    1
    """
    if high > 0:
        return max(0, low)
    return high


def calc_high_x(low: float, high: float) -> float:
    """
    From the low and high value of a parameter,
    calculates the x-value (length of bar) of the bar
    visualizing high values.
    >>> calc_high_x(-1, 1)
    1
    >>> calc_high_x(0.5, 1)
    0.5
    >>> calc_high_x(-4, -3)
    0
    """
    if high > 0:
        base = max(0, low)
        return high - base
    return 0.0


def calc_low_x(low: float, high: float) -> float:
    """
    From the low and high value of a parameter,
    calculates the x-value (length of bar) of the bar
    visualizing low values.
    >>> calc_low_x(-1, 1)
    -1
    >>> calc_low_x(-1, -0.5)
    -0.5
    >>> calc_low_x(1, 3)
    0
    """
    if low < 0:
        base = min(0, high)
        return low - base
    return 0.0


def printable_int_list(integer_list: Optional[List[int]]) -> str:
    """Creates a string out of a list of integers.
    The string gives a range x-y if all the integers between and
    including x and y are in the list, otherwise separated by commas.
    Example: the list [0, 1, 2, 4, 5, 8, 10] becomes '0-2, 4-5, 8, 10'
    If the input is `None` or the list is empty, the string 'None' is returned.
    """
    if not integer_list:
        return "None"

    sorted_list = sorted(integer_list)
    prev_number = sorted_list[0]
    string = str(prev_number)

    for next_number in sorted_list[1:]:
        if next_number > prev_number + 1:
            string += (
                f"{prev_number}" if string.endswith("-") else ""
            ) + f", {next_number}"
        elif not string.endswith("-"):
            string += "-"
        prev_number = next_number

    if not string.endswith(f", {prev_number}") and string != str(prev_number):
        string += str(prev_number)
    return string



########################################
#
#  Copyright (C) 2020-     Equinor ASA
#
########################################

import sys
from typing import Dict, List, Any
import warnings

import pandas as pd

# opm and ecl2df are only available for Linux,
# hence, ignore any import exception here to make
# it still possible to use the PvtPlugin on
# machines with other OSes.
#
# NOTE: Functions in this file cannot be used
#       on non-Linux OSes.
try:
    import ecl2df
    from opm.io.ecl import EclFile
except ImportError:
    pass

from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore

from .opm_init_io import (
    Oil,
    Gas,
    Water,
    DryGas,
    WetGas,
    LiveOil,
    DeadOil,
    VariateAndValues,
)
from .fmu_input import load_ensemble_set, load_csv


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def filter_pvt_data_frame(
    data_frame: pd.DataFrame, drop_ensemble_duplicates: bool = False
) -> pd.DataFrame:

    data_frame = data_frame.rename(str.upper, axis="columns").rename(
        columns={
            "TYPE": "KEYWORD",
            "RS": "GOR",
            "RSO": "GOR",
            "R": "GOR",
            "RV": "OGR",
        }
    )
    data_frame = data_frame.fillna(0)
    if "GOR" in data_frame.columns and "OGR" in data_frame.columns:
        data_frame["RATIO"] = data_frame["GOR"] + data_frame["OGR"]
    elif "GOR" in data_frame.columns:
        data_frame["RATIO"] = data_frame["GOR"]
    elif "OGR" in data_frame.columns:
        data_frame["RATIO"] = data_frame["OGR"]

    columns = [
        "ENSEMBLE",
        "REAL",
        "PVTNUM",
        "KEYWORD",
        "RATIO",
        "PRESSURE",
        "VOLUMEFACTOR",
        "VISCOSITY",
    ]

    if not "RATIO" in data_frame.columns:
        raise ValueError(
            "The dataframe must contain a column for the ratio (OGR, GOR, R, RV, RS)."
        )

    data_frame = data_frame[columns]

    stored_data_frames: List[pd.DataFrame] = []
    cleaned_data_frame = data_frame.iloc[0:0]
    columns_subset = data_frame.columns.difference(["REAL", "ENSEMBLE"])

    for ens, ens_data_frame in data_frame.groupby("ENSEMBLE"):
        ens_merged_dataframe = data_frame.iloc[0:0]
        for _rno, realization_data_frame in ens_data_frame.groupby("REAL"):
            if ens_merged_dataframe.empty:
                ens_merged_dataframe = ens_merged_dataframe.append(
                    realization_data_frame
                ).reset_index(drop=True)
            else:
                ens_merged_dataframe = (
                    pd.concat([ens_merged_dataframe, realization_data_frame])
                    .drop_duplicates(
                        subset=columns_subset,
                        keep="first",
                    )
                    .reset_index(drop=True)
                )

        if ens_merged_dataframe["REAL"].nunique() > 1:
            warnings.warn(
                (
                    f"There are variations in PVT between realizations in ensemble {ens}. "
                    "This is currently not supported. Only keeping data for realization "
                    f"{ens_merged_dataframe['REAL'].iloc[0]}."
                ),
                UserWarning,
            )
            ens_merged_dataframe = ens_merged_dataframe[
                ens_merged_dataframe["REAL"] == ens_merged_dataframe["REAL"].iloc[0]
            ]
        if drop_ensemble_duplicates:
            data_frame_stored = False
            for stored_data_frame in stored_data_frames:
                if all(
                    stored_data_frame[columns_subset]
                    == ens_merged_dataframe[columns_subset]
                ):
                    data_frame_stored = True
                    break
            if data_frame_stored:
                continue
            stored_data_frames.append(ens_merged_dataframe)
        cleaned_data_frame = cleaned_data_frame.append(ens_merged_dataframe)

    return cleaned_data_frame


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def load_pvt_csv(
    ensemble_paths: dict,
    csv_file: str,
    ensemble_set_name: str = "EnsembleSet",
    drop_ensemble_duplicates: bool = False,
) -> pd.DataFrame:
    return filter_pvt_data_frame(
        load_csv(ensemble_paths, csv_file, ensemble_set_name), drop_ensemble_duplicates
    )


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def load_pvt_dataframe(
    ensemble_paths: Dict[str, str],
    ensemble_set_name: str = "EnsembleSet",
    use_init_file: bool = False,
    drop_ensemble_duplicates: bool = False,
) -> pd.DataFrame:
    # pylint: disable=too-many-statements
    # If ecl2df is not loaded, this machine is probably not
    # running Linux and the modules are not available.
    # To avoid a crash, return an empty DataFrame here.
    if "ecl2df" not in sys.modules:
        print(
            "Your operating system does not support opening and reading"
            " Eclipse files. An empty data frame will be returned and your"
            " plot will therefore not show any data points. Please specify"
            " a relative path to a PVT CSV file in your PvtPlot settings"
            " to display PVT data anyways."
        )
        return pd.DataFrame({})

    def ecl2df_pvt_data_frame(kwargs: Any) -> pd.DataFrame:
        return ecl2df.pvt.df(kwargs["realization"].get_eclfiles())

    def init_to_pvt_data_frame(kwargs: Any) -> pd.DataFrame:
        # pylint: disable-msg=too-many-locals
        # pylint: disable=too-many-nested-blocks
        # pylint: disable=too-many-branches
        ecl_init_file = EclFile(
            kwargs["realization"].get_eclfiles().get_initfile().get_filename()
        )

        oil = Oil.from_ecl_init_file(ecl_init_file)
        gas = Gas.from_ecl_init_file(ecl_init_file)
        water = Water.from_ecl_init_file(ecl_init_file)

        column_pvtnum = []
        column_oil_gas_ratio = []
        column_volume_factor = []
        column_pressure = []
        column_viscosity = []
        column_keyword = []

        if oil:
            if len(oil.tables()) > 0 and isinstance(oil.tables()[0], LiveOil):
                keyword = "PVTO"
            elif len(oil.tables()) > 0 and isinstance(oil.tables()[0], DeadOil):
                keyword = "PVDO"
            else:
                raise NotImplementedError(
                    f"Oil of type '{type(oil)}' is not supported yet."
                )

            for table_index, table in enumerate(oil.tables()):
                for outer_pair in table.get_values():
                    for inner_pair in outer_pair.y:
                        if isinstance(inner_pair, VariateAndValues):
                            column_pvtnum.append(table_index + 1)
                            column_keyword.append(keyword)
                            if isinstance(table, DeadOil):
                                column_oil_gas_ratio.append(0.0)
                            else:
                                column_oil_gas_ratio.append(outer_pair.x)
                            column_pressure.append(inner_pair.x)
                            column_volume_factor.append(
                                1.0 / inner_pair.get_values_as_floats()[0]
                            )
                            column_viscosity.append(
                                inner_pair.get_values_as_floats()[0]
                                / inner_pair.get_values_as_floats()[1]
                            )

        if gas:
            if len(gas.tables()) > 0 and isinstance(gas.tables()[0], DryGas):
                keyword = "PVDG"
            elif len(gas.tables()) > 0 and isinstance(gas.tables()[0], WetGas):
                keyword = "PVTG"
            else:
                raise NotImplementedError(
                    f"Gas of type '{type(gas)}' is not supported yet."
                )
            for table_index, table in enumerate(gas.tables()):
                for outer_pair in table.get_values():
                    for inner_pair in outer_pair.y:
                        if isinstance(inner_pair, VariateAndValues):
                            column_pvtnum.append(table_index + 1)
                            column_keyword.append(keyword)
                            if isinstance(table, DryGas):
                                column_oil_gas_ratio.append(0.0)
                                column_pressure.append(inner_pair.x)
                            else:
                                column_oil_gas_ratio.append(inner_pair.x)
                                column_pressure.append(outer_pair.x)
                            column_volume_factor.append(
                                1.0 / inner_pair.get_values_as_floats()[0]
                            )
                            column_viscosity.append(
                                inner_pair.get_values_as_floats()[0]
                                / inner_pair.get_values_as_floats()[1]
                            )
        if water:
            for table_index, table in enumerate(water.tables()):
                for outer_pair in table.get_values():
                    for inner_pair in outer_pair.y:
                        if isinstance(inner_pair, VariateAndValues):
                            column_pvtnum.append(table_index + 1)
                            column_keyword.append("PVTW")
                            column_oil_gas_ratio.append(outer_pair.x)
                            column_pressure.append(inner_pair.x)
                            column_volume_factor.append(
                                1.0 / inner_pair.get_values_as_floats()[0]
                            )
                            column_viscosity.append(
                                1.0
                                / inner_pair.get_values_as_floats()[2]
                                * inner_pair.get_values_as_floats()[0]
                            )

        data_frame = pd.DataFrame(
            {
                "PVTNUM": column_pvtnum,
                "KEYWORD": column_keyword,
                "R": column_oil_gas_ratio,
                "PRESSURE": column_pressure,
                "VOLUMEFACTOR": column_volume_factor,
                "VISCOSITY": column_viscosity,
            }
        )

        return data_frame

    return filter_pvt_data_frame(
        load_ensemble_set(ensemble_paths, ensemble_set_name).apply(
            init_to_pvt_data_frame if use_init_file else ecl2df_pvt_data_frame
        ),
        drop_ensemble_duplicates,
    )



import json

import pandas as pd
import dash
from dash.dependencies import Input, Output
import dash_html_components as html

from webviz_subsurface._private_plugins.surface_selector import SurfaceSelector

surface_context = {
    "oilthickness": {
        "names": ["lowerreek", "upperreek", "all", "midreek"],
        "dates": [
            "20030101_20010601",
            "20010601_20000101",
            "20010601",
            "20000101",
            "20030101",
            "20030101_20000101",
            "20010601_20010604",
        ],
    },
    "ds_extracted_horizons": {
        "names": ["toplowerreek", "topmidreek", "topupperreek", "baselowerreek"],
        "dates": [None],
    },
}

return_value = {
    "name": "lowerreek",
    "attr": "oilthickness",
    "date": "20030101_20010601",
}


def test_surface_selector(dash_duo: dash.testing.composite.DashComposite) -> None:

    app = dash.Dash(__name__)
    app.config.suppress_callback_exceptions = True
    realizations = pd.read_csv("tests/data/realizations.csv")
    surface_selector = SurfaceSelector(app, surface_context, realizations)

    app.layout = html.Div(
        children=[surface_selector.layout, html.Pre(id="pre", children="ok")]
    )

    @app.callback(
        Output("pre", "children"), [Input(surface_selector.storage_id, "data")]
    )
    def _test(data: str) -> str:
        return json.dumps(json.loads(data))

    dash_duo.start_server(app)

    dash_duo.wait_for_contains_text("#pre", json.dumps(return_value), timeout=4)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

TESTS_REQUIRE = [
    "selenium>=3.141",
    "pylint",
    "mypy",
    "mock",
    "black>=20.8b1",
    "bandit",
    "pytest-xdist",
]

# pylint: disable=line-too-long
setup(
    name="webviz-subsurface",
    description="Webviz config plugins for subsurface data",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/equinor/webviz-subsurface",
    author="R&T Equinor",
    packages=find_packages(exclude=["tests"]),
    package_data={
        "webviz_subsurface": [
            "_abbreviations/abbreviation_data/*.json",
            "_assets/css/*.css",
        ]
    },
    entry_points={
        "webviz_config_plugins": [
            "BhpQc = webviz_subsurface.plugins:BhpQc",
            "DiskUsage = webviz_subsurface.plugins:DiskUsage",
            "HistoryMatch = webviz_subsurface.plugins:HistoryMatch",
            "HorizonUncertaintyViewer = webviz_subsurface.plugins:HorizonUncertaintyViewer",
            "InplaceVolumes = webviz_subsurface.plugins:InplaceVolumes",
            "InplaceVolumesOneByOne = webviz_subsurface.plugins:InplaceVolumesOneByOne",
            "MorrisPlot = webviz_subsurface.plugins:MorrisPlot",
            "ParameterAnalysis = webviz_subsurface.plugins:ParameterAnalysis",
            "ParameterCorrelation = webviz_subsurface.plugins:ParameterCorrelation",
            "ParameterDistribution = webviz_subsurface.plugins:ParameterDistribution",
            "ParameterParallelCoordinates = webviz_subsurface.plugins:ParameterParallelCoordinates",
            "ParameterResponseCorrelation = webviz_subsurface.plugins:ParameterResponseCorrelation",
            "PropertyStatistics = webviz_subsurface.plugins:PropertyStatistics",
            "PvtPlot = webviz_subsurface.plugins:PvtPlot",
            "RelativePermeability = webviz_subsurface.plugins:RelativePermeability",
            "ReservoirSimulationTimeSeries = webviz_subsurface.plugins:ReservoirSimulationTimeSeries",
            "ReservoirSimulationTimeSeriesOneByOne = webviz_subsurface.plugins:ReservoirSimulationTimeSeriesOneByOne",
            "ReservoirSimulationTimeSeriesRegional = webviz_subsurface.plugins:ReservoirSimulationTimeSeriesRegional",
            "RftPlotter = webviz_subsurface.plugins:RftPlotter",
            "RunningTimeAnalysisFMU = webviz_subsurface.plugins:RunningTimeAnalysisFMU",
            "SegyViewer = webviz_subsurface.plugins:SegyViewer",
            "SubsurfaceMap = webviz_subsurface.plugins:SubsurfaceMap",
            "SurfaceViewerFMU = webviz_subsurface.plugins:SurfaceViewerFMU",
            "SurfaceWithGridCrossSection = webviz_subsurface.plugins:SurfaceWithGridCrossSection",
            "SurfaceWithSeismicCrossSection = webviz_subsurface.plugins:SurfaceWithSeismicCrossSection",
            "WellCrossSection = webviz_subsurface.plugins:WellCrossSection",
            "WellCrossSectionFMU = webviz_subsurface.plugins:WellCrossSectionFMU",
            "AssistedHistoryMatchingAnalysis = webviz_subsurface.plugins:AssistedHistoryMatchingAnalysis",
        ]
    },
    install_requires=[
        "dash>=1.11",
        "dash_bootstrap_components>=0.10.3",
        "defusedxml>=0.6.0",
        # "ecl2df>=0.6.1; sys_platform=='linux'",
        "fmu-ensemble>=1.2.3",
        # "opm>=2020.10.1; sys_platform=='linux'",
        "pandas>=0.24",
        "pillow>=6.1",
        "pyscal>=0.4.1",
        "scipy>=1.2",
        "statsmodels>=0.12.1",  # indirect dependency through https://plotly.com/python/linear-fits/
        "webviz-config>=0.0.55",
        "webviz-subsurface-components>=0.2.0",
        "xtgeo>=2.8",
    ],
    tests_require=TESTS_REQUIRE,
    extras_require={"tests": TESTS_REQUIRE},
    setup_requires=["setuptools_scm~=3.2"],
    python_requires="~=3.6",
    use_scm_version=True,
    zip_safe=False,
    project_urls={
        "Documentation": "https://equinor.github.io/webviz-subsurface",
        "Download": "https://pypi.org/project/webviz-subsurface/",
        "Source": "https://github.com/equinor/webviz-subsurface",
        "Tracker": "https://github.com/equinor/webviz-subsurface/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Environment :: Web Environment",
        "Framework :: Dash",
        "Framework :: Flask",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Visualization",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
)



from typing import Tuple, TYPE_CHECKING

import dash
from dash.dependencies import Input, Output, ALL
import dash_html_components as html
import plotly.graph_objects as go

from ..views.property_delta_view import table_view, surface_views
from ..utils.surface import surface_from_zone_prop
from ..models.surface_leaflet_model import (
    SurfaceLeafletModel,
)

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from ..property_statistics import PropertyStatistics


def property_delta_controller(parent: "PropertyStatistics", app: dash.Dash) -> None:
    @app.callback(
        Output(parent.uuid("delta-bar-graph"), "figure"),
        Output(parent.uuid("delta-table-surface-wrapper"), "children"),
        Input(parent.uuid("delta-switch-table-surface"), "value"),
        Input(parent.uuid("delta-sort"), "value"),
        Input({"id": parent.uuid("ensemble-selector"), "tab": "delta"}, "value"),
        Input({"id": parent.uuid("delta-ensemble-selector"), "tab": "delta"}, "value"),
        Input(
            {"id": parent.uuid("property-selector"), "tab": "delta"},
            "value",
        ),
        Input(
            {"id": parent.uuid("filter-selector"), "tab": "delta", "selector": ALL},
            "value",
        ),
        Input(parent.uuid("delta-bar-graph"), "clickData"),
    )
    def _update_bars(
        plot_type: str,
        sortby: str,
        ensemble: str,
        delta_ensemble: str,
        prop: str,
        selectors: list,
        clickdata: dict,
    ) -> Tuple[go.Figure, html.Div]:
        # Make bar chart
        bars = parent.pmodel.make_delta_bars(
            prop=prop,
            ensemble=ensemble,
            delta_ensemble=delta_ensemble,
            selector_values=selectors,
            aggregation=sortby,
        )

        if plot_type == "surface":
            # Get selected bar or get highest delta bar if none is selected
            label = (
                clickdata["points"][0]["y"]
                if clickdata is not None
                else bars["data"][0]["y"][-1]
            )
            # Make surface div
            wrapper = make_surfaces(
                parent=parent,
                prop=prop,
                zone=label.split(" | ")[0],
                ensemble=ensemble,
                delta_ensemble=delta_ensemble,
            )
        if plot_type == "table":
            columns, data = parent.pmodel.make_delta_table(
                prop=prop,
                ensemble=ensemble,
                delta_ensemble=delta_ensemble,
                selector_values=selectors,
                aggregation=sortby,
            )
            wrapper = table_view(data=data, columns=columns)

        return bars, wrapper


def make_surfaces(
    parent: "PropertyStatistics",
    prop: str,
    zone: str,
    ensemble: str,
    delta_ensemble: str,
    statistic: str = "mean",
) -> html.Div:

    sprop = parent.surface_renaming.get(prop, prop)
    szone = parent.surface_renaming.get(zone, zone)
    ens_surface = surface_from_zone_prop(
        parent, zone=szone, prop=sprop, ensemble=ensemble, stype=statistic
    )
    delta_ens_surface = surface_from_zone_prop(
        parent, zone=szone, prop=sprop, ensemble=delta_ensemble, stype=statistic
    )

    # Truncating surface values to highest low and lowest high for each map
    min_val = max([ens_surface.values.min(), delta_ens_surface.values.min()])
    max_val = min([ens_surface.values.max(), delta_ens_surface.values.max()])

    ens_surface.values[ens_surface.values < min_val] = min_val
    ens_surface.values[ens_surface.values > max_val] = max_val
    delta_ens_surface.values[delta_ens_surface.values < min_val] = min_val
    delta_ens_surface.values[delta_ens_surface.values > max_val] = max_val
    diff_surface = ens_surface.copy()
    diff_surface.values = ens_surface.values - delta_ens_surface.values
    return surface_views(
        ens_layer=SurfaceLeafletModel(ens_surface, name="ens_surface").layer,
        delta_ens_layer=SurfaceLeafletModel(
            delta_ens_surface, name="delta_ens_surface"
        ).layer,
        diff_layer=SurfaceLeafletModel(diff_surface, name="diff_surface").layer,
        ensemble=ensemble,
        delta_ensemble=delta_ensemble,
        parent=parent,
        prop=prop,
        zone=zone,
    )



from typing import Optional, List, Dict, Any

import numpy.ma as ma
import numpy as np
import pandas as pd
import xtgeo
from plotly.subplots import make_subplots

from .._utils.colors import hex_to_rgba


class XSectionFigure:
    """Class for plotting a cross-section of a well.

    Args:
        zmin (float): Upper level of the plot (top Y axis).
        zmax (float): Lower level of the plot (bottom Y axis).
        well (Well): XTGeo well object.
        surfaces (list): List of XTGeo RegularSurface objects
        surfacenames (list): List of surface names (str) for legend
        cube (Cube): A XTGeo Cube instance
        grid (Grid): A XTGeo Grid instance
        gridproperty (GridProperty): A XTGeo GridProperty instance

    """

    # pylint: disable=too-many-instance-attributes, too-many-arguments
    def __init__(
        self,
        zmin: Optional[float] = None,
        zmax: Optional[float] = None,
        well: Optional[xtgeo.Well] = None,
        surfaces: Optional[List[xtgeo.RegularSurface]] = None,
        sampling: int = 5,
        nextend: int = 5,
        zonelogshift: int = 0,
        surfacenames: Optional[List[str]] = None,
        surfacecolors: Optional[List[str]] = None,
        cube: Optional[xtgeo.Cube] = None,
        grid: Optional[xtgeo.Grid] = None,
        gridproperty: Optional[xtgeo.GridProperty] = None,
        zunit: str = "",
        show_marginal: bool = False,
    ) -> None:

        self._data: List[Any] = []
        self._figure = make_subplots(
            rows=2 if show_marginal else 1,
            cols=1,
            shared_xaxes=True,
            vertical_spacing=0,
            row_heights=[0.05, 0.95] if show_marginal else [1],
        )
        if well:
            self._zmin = zmin if zmin else well.dataframe["Z_TVDSS"].min()
            self._zmax = zmax if zmax else well.dataframe["Z_TVDSS"].max()
        self._well = well
        self._nextend = nextend
        self._sampling = sampling
        self._surfaces = surfaces
        self._surfacenames = surfacenames
        self._surfacecolors = (
            surfacecolors
            if surfacecolors is not None
            else [
                "#1f77b4",  # muted blue
                "#ff7f0e",  # safety orange
                "#2ca02c",  # cooked asparagus green
                "#d62728",  # brick red
                "#9467bd",  # muted purple
                "#8c564b",  # chestnut brown
                "#e377c2",  # raspberry yogurt pink
                "#7f7f7f",  # middle gray
                "#bcbd22",  # curry yellow-green
                "#17becf",  # blue-teal
            ]
        )
        self.zunit = zunit
        self.show_marginal = show_marginal
        self.main_trace_row = 2 if show_marginal else 1
        self._cube = cube
        self._grid = grid
        self._gridproperty = gridproperty
        self._zonelogshift = zonelogshift

        self._fence = None

    @property
    def data(self) -> Any:
        """The Plotly figure traces"""
        # pylint: disable=line-too-long
        # TODO: According to
        # https://plotly.com/python-api-reference/generated/plotly.graph_objects.Figure.html
        # this can return very different types.
        # pylint: enable=line-too-long
        return self._figure.data

    @property
    def layout(self) -> Dict[str, Any]:
        """The Plotly figure layout"""
        _layout = self._figure.layout
        yax = "yaxis2" if self.show_marginal else "yaxis"
        _layout.update(
            {
                "height": 800,
                # "margin": {"t":300},
                "xaxis2": {
                    "showgrid": False,
                    "title": "Distance from well",
                    "zeroline": False,
                },
                yax: {
                    "range": [self._zmax, self._zmin],
                    "showgrid": False,
                    "title": "Depth",
                    "zeroline": False,
                },
                "legend": {"traceorder": "normal"},
                "hovermode": "x",
            }
        )

        return _layout

    @property
    def fence(self) -> np.ndarray:
        """Set or get the fence spesification"""
        if self._fence is None:
            if self._well is not None:
                wfence = self._well.get_fence_polyline(
                    sampling=self._sampling, nextend=self._nextend, tvdmin=self._zmin
                )
                self._fence = wfence

                if wfence is False:
                    self._fence = None
            else:
                raise ValueError("Input well is None")  # should be more flexible
        return self._fence

    def plot_well(
        self,
        zonelogname: Optional[str] = "ZONELOG",
        facieslogname: Optional[str] = None,
        marginal_log: Optional[str] = None,
        zonemin: int = 0,
    ) -> None:
        """Input an XTGeo Well object and plot it."""
        well = self._well

        # reduce the well data by Pandas operations
        dfr = well.dataframe
        well.dataframe = dfr[dfr["Z_TVDSS"] > self._zmin]

        # Create a relative XYLENGTH vector (0.0 where well starts)
        well.create_relative_hlen()

        dfr = well.dataframe

        # get the well trajectory (numpies) as copy
        zvals = dfr["Z_TVDSS"].values.copy()
        hvals = dfr["R_HLEN"].values.copy()
        self._plot_well_traj(zvals, hvals)
        if zonelogname:
            self._plot_well_zlog(dfr, zvals, hvals, zonelogname, zonemin)

        # plot the facies, if any, behind the trajectory; ie. first or second
        if facieslogname:
            self._plot_well_faclog(dfr, zvals, hvals, facieslogname)
        if marginal_log:
            self._plot_marginal_log(dfr, zvals, hvals, marginal_log)

    def _plot_well_traj(self, zvals: np.ndarray, hvals: np.ndarray) -> None:
        """Plot the trajectory as a black line"""

        zvals_copy = ma.masked_where(zvals < self._zmin, zvals)
        hvals_copy = ma.masked_where(zvals < self._zmin, hvals)

        self._figure.add_trace(
            {
                "x": hvals_copy,
                "y": zvals_copy,
                "name": self._well.name,
                "marker": {"color": "black"},
            },
            self.main_trace_row,
            1,
        )

        # ax.plot(hvals_copy, zvals_copy, linewidth=6, c="black")

    # pylint: disable-too-many-locals
    def _plot_well_zlog(
        self,
        df: pd.DataFrame,
        zvals: np.ndarray,
        hvals: np.ndarray,
        zonelogname: str,
        zomin: int = -999,
    ) -> None:
        """Plot the zone log as colored segments."""

        if zonelogname not in df.columns:
            return
        zonevals = df[zonelogname].values
        zomin = (
            zomin if zomin >= int(df[zonelogname].min()) else int(df[zonelogname].min())
        )
        zomax = int(df[zonelogname].max())

        # To prevent gaps in the zonelog it is necessary to duplicate each zone transition
        zone_transitions = np.where(zonevals[:-1] != zonevals[1:])
        for transition in zone_transitions:
            try:
                zvals = np.insert(zvals, transition, zvals[transition + 1])
                hvals = np.insert(hvals, transition, hvals[transition + 1])
                zonevals = np.insert(zonevals, transition, zonevals[transition])
            except IndexError:
                pass

        for i, zone in enumerate(range(zomin, zomax + 1)):
            zvals_copy = ma.masked_where(zonevals != zone, zvals)
            hvals_copy = ma.masked_where(zonevals != zone, hvals)
            color = self._surfacecolors[i % len(self._surfacecolors)]
            self._figure.add_trace(
                {  # pylint: disable=no-member
                    "x": hvals_copy.compressed(),
                    # pylint: disable=no-member
                    "y": zvals_copy.compressed(),
                    "line": {"width": 10, "color": color},
                    "fillcolor": color,
                    "marker": {"opacity": 0.5},
                    "showlegend": False,
                    "name": f"Zone: {zone}",
                },
                self.main_trace_row,
                1,
            )

    def _plot_well_faclog(
        self, df: pd.DataFrame, zvals: np.ndarray, hvals: np.ndarray, facieslogname: str
    ) -> None:
        """Plot the facies log as colored segments.

        Args:
            df (dataframe): The Well dataframe.
            zvals (ndarray): The numpy Z TVD array.
            hvals (ndarray): The numpy Length  array.
            facieslogname (str): name of the facies log.
        """

        if facieslogname not in df.columns:
            return

        frecord = self._well.get_logrecord(facieslogname)
        frecord = {val: fname for val, fname in frecord.items() if val >= 0}

        faciesvalues = df[facieslogname].values

        for fcc in frecord:

            zvals_copy = ma.masked_where(faciesvalues != fcc, zvals)
            hvals_copy = ma.masked_where(faciesvalues != fcc, hvals)

            self._figure.add_trace(
                {
                    "x": hvals_copy,
                    "y": zvals_copy,
                    "line": {"width": 5},
                    "connectgaps": True,
                },
                self.main_trace_row,
                1,
            )

        # self._drawlegend(ax, bba, title="Facies")

    def _plot_marginal_log(
        self,
        df: pd.DataFrame,
        zvals: np.ndarray,
        hvals: np.ndarray,
        logname: str,
        row: int = 1,
    ) -> None:
        """Plot a marginal log above main plot"""
        if logname in df.columns:
            self._figure.add_trace(
                {
                    "mode": "lines",
                    "fill": "tozeroy",
                    "x": hvals,
                    "y": df[logname],
                    "name": logname,
                    "hovertext": [
                        f"TVD: {zvals[i]}<br>" f"{logname}: {list(df[logname])[i]}<br>"
                        for i, _ in enumerate(hvals)
                    ],
                    "hoverinfo": "text",
                    "line": {"color": self._surfacecolors[0]},
                    "fillcolor": self._surfacecolors[0],
                },
                row,
                1,
            )

    # pylint: disable=too-many-locals, unused-argument
    def plot_cube(
        self,
        cube: xtgeo.Cube = None,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
        alpha: float = 0.7,
        interpolation: str = "gaussian",
        sampling: str = "nearest",
        name: str = "seismic",
    ) -> None:
        """Plot a cube backdrop.

        Args:
            vmin (float): Minimum value in plot.
            vmax (float); Maximum value in plot
            alpha (float): Alpah blending number beween 0 and 1.
            interpolation (str): Interpolation for plotting, cf. matplotlib
                documentation on this. Also gaussianN is allowed, where
                N = 1..9.
            sampling (str): 'nearest' (default) or 'trilinear' (more precise)

        Raises:
            ValueError: No cube is loaded

        """
        if cube:
            self._cube = cube
        if self._cube is None:
            raise ValueError("Ask for plot cube, but noe cube is loaded")

        zinc = self._cube.zinc / 2.0

        zvalsv = self._cube.get_randomline(
            self.fence,
            zmin=self._zmin,
            zmax=self._zmax,
            zincrement=zinc,
            sampling=sampling,
        )

        xmin, xmax, ymin, ymax, arr = zvalsv
        x_inc = (xmax - xmin) / arr.shape[1]
        y_inc = (ymax - ymin) / arr.shape[0]
        # if vmin is not None or vmax is not None:
        #     arr = np.clip(arr, vmin, vmax)

        self._figure.add_trace(
            {
                "type": "heatmap",
                "z": arr,
                "x0": xmin,
                # "xmax": xmax,
                "dx": x_inc,
                "y0": ymin,
                # "ymax": ymax,
                "dy": y_inc,
                "zsmooth": "best",
                "showscale": False,
                "name": name
                # "colorscale": colors,
            },
            self.main_trace_row,
            1,
        )

    # pylint: disable=too-many-locals, unused-argument
    def plot_grid3d(
        self,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
        alpha: float = 0.7,
    ) -> None:
        """Plot a sampled grid with gridproperty backdrop.

        Args:
            vmin (float): Minimum value in plot.
            vmax (float); Maximum value in plot
            alpha (float): Alpha blending number beween 0 and 1.

        Raises:
            ValueError: No grid or gridproperty is loaded

        """
        if self._grid is None or self._gridproperty is None:
            raise ValueError("Ask for plot of grid, but no grid is loaded")

        zinc = 0.5  # tmp

        zvalsv = self._grid.get_randomline(
            self.fence,
            self._gridproperty,
            zmin=self._zmin,
            zmax=self._zmax,
            zincrement=zinc,
        )

        xmin, xmax, ymin, ymax, arr = zvalsv
        x_inc = (xmax - xmin) / arr.shape[1]
        y_inc = (ymax - ymin) / arr.shape[0]
        # if vmin is not None or vmax is not None:
        #     arr = np.clip(arr, vmin, vmax)

        self._figure.add_trace(
            {
                "type": "heatmap",
                "z": arr,
                "x0": xmin,
                "xmax": xmax,
                "dx": x_inc,
                "y0": ymin,
                "ymax": ymax,
                "dy": y_inc,
                "zsmooth": "best",
                "showscale": False,
                # "colorscale": colors,
            },
            self.main_trace_row,
            1,
        )

    def plot_surfaces(
        self,
        fill: bool = False,
        surfaces: List[xtgeo.RegularSurface] = None,
        surfacenames: Optional[List[str]] = None,
    ) -> None:  # pylint: disable=too-many-branches, too-many-statements

        """Input a surface list (ordered from top to base) , and plot them."""

        # ax, bba = self._currentax(axisname=axisname)

        # either use surfaces from __init__, or override with surfaces
        # specified here

        if surfaces and surfacenames and len(surfaces) == len(surfacenames):
            for i, surface in enumerate(surfaces):

                hfence1 = surface.get_randomline(self.fence).copy()
                self._figure.add_trace(
                    {
                        "type": "scatter",
                        "mode": "lines",
                        "fill": "tonexty" if i != 0 and fill else None,
                        "y": hfence1[:, 1],
                        "x": hfence1[:, 0],
                        "name": surfacenames[i],
                        # "marker": {"color": s_color},
                    },
                    self.main_trace_row,
                    1,
                )

    def plot_statistical_surface(
        self,
        statistical_surfaces: Dict[str, xtgeo.RegularSurface],
        name: str,
        fill: bool = False,
    ) -> None:
        """Plot statistical surfaces (p10/p90/min/max/mean) as fanchart

        Args:
            statistical_surfaces: Dict of xtgeo surfaces
            name: Name of surface (e.g. horizon name)
            fill: Toggles fill between surfaces
        """
        if self._surfacecolors and self._surfacenames:
            color = self._surfacecolors[
                self._surfacenames.index(name) % len(self._surfacecolors)
            ]
        else:
            color = "rgba(0, 0, 0, 1.0)"
            # Either self._surfacecolors or self._surfacenames not set.
            assert False
        fill_color = hex_to_rgba(color, 0.3)
        line_color = hex_to_rgba(color, 1)

        # Extract surface values along well fence
        x_values = statistical_surfaces["mean"].get_randomline(self.fence).copy()[:, 0]
        stat = {
            key: statistical_surfaces[key].get_randomline(self.fence).copy()[:, 1]
            for key in ["maximum", "minimum", "p90", "p10", "mean", "stddev"]
        }

        # Maximum trace (contains hoverinfo)
        self._figure.add_trace(
            {
                "name": name,
                "y": stat["maximum"],
                "x": x_values,
                "mode": "lines",
                "hovertext": [
                    f"Minimum: {stat['minimum'][i]:.2f} {self.zunit}<br>"
                    f"P10: {stat['p10'][i]:.2f} {self.zunit}<br>"
                    f"Mean: {stat['mean'][i]:.2f} {self.zunit}<br>"
                    f"P90: {stat['p90'][i]:.2f} {self.zunit}<br>"
                    f"Maximum: {stat['maximum'][i]:.2f} {self.zunit}<br>"
                    f"Std.Dev: {stat['stddev'][i]:.2f} {self.zunit}"
                    for i, _ in enumerate(x_values)
                ],
                "hoverinfo": "text+name",
                "line": {"width": 0 if fill else 1, "color": line_color},
                "legendgroup": name,
                "showlegend": False,
            },
            self.main_trace_row,
            1,
        )
        # P10 trace
        self._figure.add_trace(
            {
                "name": name,
                "y": stat["p10"],
                "x": x_values,
                "mode": "lines",
                "hoverinfo": "skip",
                "fill": "tonexty" if fill else None,
                "fillcolor": fill_color,
                "line": {"width": 0 if fill else 1, "color": line_color},
                "legendgroup": name,
                "showlegend": False,
            },
            self.main_trace_row,
            1,
        )
        # Mean trace
        self._figure.add_trace(
            {
                "name": name,
                "y": stat["mean"],
                "x": x_values,
                "mode": "lines",
                "hoverinfo": "skip",
                "fill": "tonexty" if fill else None,
                "fillcolor": fill_color,
                "line": {"color": line_color},
                "legendgroup": name,
                "showlegend": True,
            },
            self.main_trace_row,
            1,
        )
        # P90 trace
        self._figure.add_trace(
            {
                "name": name,
                "y": stat["p90"],
                "x": x_values,
                "mode": "lines",
                "hoverinfo": "skip",
                "fill": "tonexty" if fill else None,
                "fillcolor": fill_color,
                "line": {"width": 0 if fill else 1, "color": line_color},
                "legendgroup": name,
                "showlegend": False,
            },
            self.main_trace_row,
            1,
        )
        # Minimum trace
        self._figure.add_trace(
            {
                "name": name,
                "y": stat["minimum"],
                "x": x_values,
                "mode": "lines",
                "hoverinfo": "skip",
                "fill": "tonexty" if fill else None,
                "fillcolor": fill_color,
                "line": {"width": 0 if fill else 1, "color": line_color},
                "legendgroup": name,
                "showlegend": False,
            },
            self.main_trace_row,
            1,
        )



from typing import List, Optional, Dict, Any

import pandas as pd
import numpy as np

from ._processing import interpolate_depth, filter_frame
from ..._utils.colors import hex_to_rgba


class FormationFigure:
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        well: str,
        ertdf: pd.DataFrame,
        enscolors: dict,
        depth_option: str,
        date: str,
        ensembles: List[str],
        simdf: Optional[pd.DataFrame] = None,
        obsdf: Optional[pd.DataFrame] = None,
    ) -> None:
        self.well = well
        self.simdf = (
            filter_frame(simdf, {"WELL": well, "DATE": date, "ENSEMBLE": ensembles})
            if simdf is not None
            else None
        )
        self.obsdf = (
            filter_frame(obsdf, {"WELL": well, "DATE": date})
            if obsdf is not None
            else None
        )
        self.ertdf = filter_frame(
            ertdf, {"WELL": well, "DATE": date, "ENSEMBLE": ensembles}
        )
        self.depth_option = depth_option
        self.enscolors = enscolors
        self.set_depth_columns()

        self.traces: List[Dict[str, Any]] = []
        self._layout = {
            "yaxis": {"autorange": "reversed", "title": "Depth", "showgrid": False},
            "xaxis": {
                "title": "Pressure",
                "showgrid": False,
                "range": [
                    (self.pressure_range[0] - self.xaxis_extension),
                    (self.pressure_range[1] + self.xaxis_extension),
                ],
            },
            "height": 800,
            "legend": {"orientation": "h"},
            "margin": {"t": 50},
            "hovermode": "closest",
            "uirevision": f"{well}{depth_option}",
        }

    @property
    def layout(self) -> Dict[str, Any]:
        return self._layout

    @property
    def xaxis_extension(self) -> float:
        min_x, max_x = self.pressure_range
        return (max_x - min_x) * 0.5

    @property
    def pressure_range(self) -> List[float]:
        min_sim = (
            self.ertdf["SIMULATED"].min()
            if self.use_ertdf
            else self.simdf["PRESSURE"].min()
        )
        max_sim = (
            self.ertdf["SIMULATED"].max()
            if self.use_ertdf
            else self.simdf["PRESSURE"].max()
        )
        min_obs = (self.ertdf["OBSERVED"] - self.ertdf["OBSERVED_ERR"]).min()
        max_obs = (self.ertdf["OBSERVED"] + self.ertdf["OBSERVED_ERR"]).max()

        return [
            min(
                min_sim,
                min_obs,
            ),
            max(
                max_sim,
                max_obs,
            ),
        ]

    @property
    def simdf_has_md(self) -> bool:
        if (
            self.simdf is not None
            and "CONMD" in self.simdf
            and len(self.simdf["CONMD"].unique()) == len(self.simdf["DEPTH"].unique())
        ):
            return True
        return False

    @property
    def use_ertdf(self) -> bool:
        return self.simdf is None or (
            self.depth_option == "MD" and not self.simdf_has_md
        )

    def set_depth_columns(self) -> None:
        """Set depth columns (md vs tvd)"""
        self.ertdf["DEPTH"] = self.ertdf["TVD"]

        if self.depth_option == "MD":
            self.ertdf["DEPTH"] = self.ertdf["MD"]
            if self.simdf_has_md:
                self.simdf["DEPTH"] = self.simdf["CONMD"]
            if self.obsdf is not None and "MD" in self.obsdf:
                self.obsdf["DEPTH"] = self.obsdf["MD"]

    def add_formation(self, df: pd.DataFrame) -> None:
        """Plot zonation"""
        formation_names = []
        formation_colors = [
            f"hsl({210*(1-x) + 202*x}, {100*(1-x) + 56*x}%, {98*(1-x) + 62*x}%)"
            for x in np.linspace(0, 1, len(df["ZONE"].unique()))
        ]

        top_col = "TOP_TVD" if self.depth_option == "TVD" else "TOP_MD"
        base_col = "BASE_TVD" if self.depth_option == "TVD" else "BASE_MD"

        df = filter_frame(df, {"WELL": self.well})
        df = df[df["TOP_MD"] != df["BASE_MD"]]

        for (_index, row) in df.iterrows():
            if pd.isnull(row[base_col]):
                raise KeyError(f"{base_col} has missing information")

            formation_names.append(
                {
                    "xref": "paper",
                    "yref": "y",
                    "x0": 0,
                    "x1": 1,
                    "name": row["ZONE"],
                    "y0": row[top_col],
                    "y1": row[base_col],
                    "linecolor": formation_colors[
                        list(df["ZONE"].unique()).index(row["ZONE"])
                    ],
                    "fillcolor": formation_colors[
                        list(df["ZONE"].unique()).index(row["ZONE"])
                    ],
                    "type": "rect",
                    "layer": "below",
                }
            )
            # Add formation names
            self.traces.append(
                {
                    "showlegend": _index - 1 == df.index.min(),
                    "name": "Formations",
                    "legendgroup": "name",
                    "type": "scatter",
                    "y": [abs((row[base_col] + row[top_col]) / 2)],
                    "x": [self.pressure_range[1] + (self.xaxis_extension / 1.2)],
                    "text": f"<b>{row['ZONE']}</b>",
                    "textfont": {"size": 14, "color": "#E50000"},
                    "textposition": "middle left",
                    "mode": "text",
                    "hoverinfo": "skip",
                }
            )
        self._layout.update({"shapes": formation_names})

    def add_additional_observations(self) -> None:
        if self.obsdf is not None:
            self.traces.append(
                {
                    "x": self.obsdf["PRESSURE"],
                    "y": self.obsdf["DEPTH"],
                    "type": "scatter",
                    "mode": "markers",
                    "name": "Observations",
                    "marker": {"color": "Black", "size": 20},
                }
            )

    def add_ert_observed(self) -> None:
        df = self.ertdf.copy()
        df = filter_frame(
            df,
            {
                "REAL": df["REAL"].unique()[0],
                "ENSEMBLE": df["ENSEMBLE"].unique()[0],
            },
        )
        self.traces.append(
            {
                "x": df["OBSERVED"],
                "y": df["DEPTH"],
                "type": "scatter",
                "mode": "markers",
                "name": "Ert observations",
                "marker": {
                    "color": "#2584DE",
                    "size": 20,
                },
                "error_x": {
                    "type": "data",
                    "array": df["OBSERVED_ERR"],
                    "visible": True,
                    "thickness": 6,
                },
            }
        )

    def add_simulated_lines(self, linetype: str) -> None:
        if self.use_ertdf:
            for ensemble, ensdf in self.ertdf.groupby("ENSEMBLE"):
                self.traces.append(
                    {
                        "x": ensdf["SIMULATED"],
                        "y": ensdf["DEPTH"],
                        "type": "scatter",
                        "mode": "markers",
                        "name": ensemble,
                        "marker": {
                            "color": self.enscolors[ensemble],
                            "size": 20,
                            "line": {"width": 1, "color": "grey"},
                        },
                    }
                )
        else:
            if linetype == "realization":
                for ensemble, ensdf in self.simdf.groupby("ENSEMBLE"):
                    for i, (real, realdf) in enumerate(ensdf.groupby("REAL")):
                        self.traces.append(
                            {
                                "x": realdf["PRESSURE"],
                                "y": realdf["DEPTH"],
                                "hoverinfo": "y+x+text",
                                "hovertext": f"Realization: {real}, Ensemble: {ensemble}",
                                "type": "scatter",
                                "mode": "lines",
                                "line": {"color": self.enscolors[ensemble]},
                                "name": ensemble,
                                "showlegend": i == 0,
                                "legendgroup": ensemble,
                            }
                        )
            if linetype == "fanchart":
                for ensemble, ensdf in self.simdf.groupby("ENSEMBLE"):
                    quantiles = [10, 90]
                    dframe = (
                        interpolate_depth(ensdf).drop(columns=["REAL"]).groupby("DEPTH")
                    )
                    # Build a dictionary of dataframes to be concatenated
                    dframes = {}
                    dframes["mean"] = dframe.mean()
                    for quantile in quantiles:
                        quantile_str = "p" + str(quantile)
                        dframes[quantile_str] = dframe.quantile(q=quantile / 100.0)
                    dframes["maximum"] = dframe.max()
                    dframes["minimum"] = dframe.min()
                    self.traces.extend(
                        add_fanchart_traces(
                            pd.concat(dframes, names=["STATISTIC"], sort=False)[
                                "PRESSURE"
                            ],
                            self.enscolors[ensemble],
                            ensemble,
                        )
                    )


def add_fanchart_traces(
    vector_stats: pd.DataFrame, color: str, legend_group: str
) -> List[Dict[str, Any]]:
    """Renders a fanchart for an ensemble vector"""
    fill_color = hex_to_rgba(color, 0.3)
    line_color = hex_to_rgba(color, 1)
    return [
        {
            "name": legend_group,
            "hovertext": "Maximum",
            "y": vector_stats["maximum"].index.tolist(),
            "x": vector_stats["maximum"].values,
            "mode": "lines",
            "line": {"width": 0, "color": line_color},
            "legendgroup": legend_group,
            "showlegend": False,
        },
        {
            "name": legend_group,
            "hovertext": "P10",
            "y": vector_stats["p10"].index.tolist(),
            "x": vector_stats["p10"].values,
            "mode": "lines",
            "fill": "tonexty",
            "fillcolor": fill_color,
            "line": {"width": 0, "color": line_color},
            "legendgroup": legend_group,
            "showlegend": False,
        },
        {
            "name": legend_group,
            "hovertext": "Mean",
            "y": vector_stats["mean"].index.tolist(),
            "x": vector_stats["mean"].values,
            "mode": "lines",
            "fill": "tonexty",
            "fillcolor": fill_color,
            "line": {"color": line_color},
            "legendgroup": legend_group,
            "showlegend": True,
        },
        {
            "name": legend_group,
            "hovertext": "P90",
            "y": vector_stats["p90"].index.tolist(),
            "x": vector_stats["p90"].values,
            "mode": "lines",
            "fill": "tonexty",
            "fillcolor": fill_color,
            "line": {"width": 0, "color": line_color},
            "legendgroup": legend_group,
            "showlegend": False,
        },
        {
            "name": legend_group,
            "hovertext": "Minimum",
            "y": vector_stats["minimum"].index.tolist(),
            "x": vector_stats["minimum"].values,
            "mode": "lines",
            "fill": "tonexty",
            "fillcolor": fill_color,
            "line": {"width": 0, "color": line_color},
            "legendgroup": legend_group,
            "showlegend": False,
        },
    ]



from typing import List, Tuple, Union, Callable, Iterable
import json
from uuid import uuid4
from pathlib import Path

import numpy as np
import pandas as pd
import dash
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State
from dash_table import DataTable
import dash_html_components as html
import dash_core_components as dcc
import webviz_core_components as wcc
from webviz_config import WebvizPluginABC
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore

from .._private_plugins.tornado_plot import TornadoPlot
from .._datainput.inplace_volumes import extract_volumes
from .._datainput.fmu_input import get_realizations, find_sens_type
from .._abbreviations.volume_terminology import volume_description, volume_unit
from .._abbreviations.number_formatting import table_statistics_base


class InplaceVolumesOneByOne(WebvizPluginABC):
    # pylint: disable=too-many-instance-attributes
    """Visualizes inplace volumetrics related to a FMU ensemble with a design matrix.

Input can be given either as an aggregated `csv` file for volumes and sensitivity information,
or as ensemble name(s) defined in `shared_settings` and volumetric `csv` files
stored per realization.

---

* **`csvfile_vol`:** Aggregated csvfile for volumes with `REAL`, `ENSEMBLE` and `SOURCE` columns.
* **`csvfile_parameters`:** Aggregated csvfile of parameters for sensitivity information with \
  `REAL`, `ENSEMBLE`, `SENSNAME` and `SENSCASE` columns.
* **`ensembles`:** Which ensembles in `shared_settings` to visualize (not to be used with \
  `csvfile_vol` and `csvfile_parameters`).
* **`volfiles`:**  Key/value pair of csv files when using `ensembles`. \
  E.g. `{geogrid: geogrid--oil.csv}`.
* **`volfolder`:** Optional local folder for the `volfiles`.
* **`response`:** Optional volume response to visualize initially.

---
?> The input files must follow FMU standards.


**Volumetric input**

* [Example of an aggregated file for `csvfile_vol`](https://github.com/equinor/\
webviz-subsurface-testdata/blob/master/aggregated_data/volumes.csv).

* [Example of a file per realization that can be used with `ensembles` and `volfiles`]\
(https://github.com/equinor/webviz-subsurface-testdata/blob/master/reek_history_match/\
realization-0/iter-0/share/results/volumes/geogrid--oil.csv).

The following columns will be used as available filters, if present:

* `ZONE`
* `REGION`
* `FACIES`
* `LICENSE`
* `SOURCE` (relevant if calculations are done for multiple grids)


Remaining columns are seen as volumetric responses.

All names are allowed (except those mentioned above, in addition to `REAL` and `ENSEMBLE`), \
but the following responses are given more descriptive names automatically:

* `BULK_OIL`: Bulk Volume (Oil)
* `NET_OIL`: Net Volume (Oil)
* `PORE_OIL`: Pore Volume (Oil)
* `HCPV_OIL`: Hydro Carbon Pore Volume (Oil)
* `STOIIP_OIL`: Stock Tank Oil Initially In Place
* `BULK_GAS`: Bulk Volume (Gas)
* `NET_GAS`: Net Volume (Gas)
* `PORV_GAS`: Pore Volume (Gas)
* `HCPV_GAS`: Hydro Carbon Pore Volume (Gas)
* `GIIP_GAS`: Gas Initially In Place
* `RECOVERABLE_OIL`: Recoverable Volume (Oil)
* `RECOVERABLE_GAS`: Recoverable Volume (Gas)

**Sensitivity input**

The sensitivity information is extracted automatically if `ensembles` is given as input,
as long as `SENSCASE` and `SENSNAME` is found in `parameters.txt`.

An example of an aggregated file to use with `csvfile_parameters`
[can be found here](https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
aggregated_data/parameters.csv)
"""

    FILTERS = ["ZONE", "REGION", "FACIES", "LICENSE"]

    TABLE_STATISTICS: List[Tuple[str, dict]] = [
        ("Sens Name", {}),
        ("Sens Case", {}),
    ] + table_statistics_base()

    ENSEMBLE_COLUMNS = [
        "REAL",
        "ENSEMBLE",
        "SOURCE",
        "SENSCASE",
        "SENSNAME",
        "SENSTYPE",
        "RUNPATH",
    ]

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        app: dash.Dash,
        csvfile_vol: Path = None,
        csvfile_parameters: Path = None,
        ensembles: list = None,
        volfiles: dict = None,
        volfolder: str = "share/results/volumes",
        response: str = "STOIIP_OIL",
    ):

        super().__init__()

        self.csvfile_vol = csvfile_vol if csvfile_vol else None
        self.csvfile_parameters = csvfile_parameters if csvfile_parameters else None

        if csvfile_vol and ensembles:
            raise ValueError(
                'Incorrent arguments. Either provide a "csvfile_vol" and "csvfile_parameters" or '
                '"ensembles" and "volfiles"'
            )
        if csvfile_vol and csvfile_parameters:
            volumes = read_csv(csvfile_vol)
            parameters = read_csv(csvfile_parameters)
            parameters["SENSTYPE"] = parameters.apply(
                lambda row: find_sens_type(row.SENSCASE), axis=1
            )

        elif ensembles and volfiles:
            self.ens_paths = {
                ens: app.webviz_settings["shared_settings"]["scratch_ensembles"][ens]
                for ens in ensembles
            }
            self.volfiles = volfiles
            self.volfolder = volfolder
            # Extract volumetric dataframe
            volumes = extract_volumes(self.ens_paths, self.volfolder, self.volfiles)
            # Extract realizations and sensitivity information
            parameters = get_realizations(
                ensemble_paths=self.ens_paths, ensemble_set_name="EnsembleSet"
            )

        else:
            raise ValueError(
                'Incorrent arguments. Either provide a "csvfile_vol" and "csvfile_parameters" or '
                '"ensembles" and "volfiles"'
            )
        self.initial_response = response

        # Merge into one dataframe
        # (TODO: Should raise error if not all ensembles have sensitivity data)
        self.volumes = pd.merge(volumes, parameters, on=["ENSEMBLE", "REAL"])

        # Initialize a tornado plot. Data is added in callback
        self.tornadoplot = TornadoPlot(app, parameters, allow_click=True)
        self.uid = uuid4()
        self.selectors_id = {x: self.uuid(x) for x in self.selectors}
        self.theme = app.webviz_settings["theme"]
        self.set_callbacks(app)

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return (
            [
                (
                    read_csv,
                    [
                        {"csv_file": self.csvfile_vol},
                        {"csv_file": self.csvfile_parameters},
                    ],
                )
            ]
            if self.csvfile_vol and self.csvfile_parameters
            else [
                (
                    extract_volumes,
                    [
                        {
                            "ensemble_paths": self.ens_paths,
                            "volfolder": self.volfolder,
                            "volfiles": self.volfiles,
                        }
                    ],
                ),
                (
                    get_realizations,
                    [
                        {
                            "ensemble_paths": self.ens_paths,
                            "ensemble_set_name": "EnsembleSet",
                        }
                    ],
                ),
            ]
        )

    def selector(self, label: str, id_name: str, column: str) -> html.Div:
        return html.Div(
            children=html.Label(
                children=[
                    html.Span(f"{label}:", style={"font-weight": "bold"}),
                    dcc.Dropdown(
                        id=self.uuid(id_name),
                        options=[
                            {"label": i, "value": i}
                            for i in list(self.volumes[column].unique())
                        ],
                        clearable=False,
                        value=list(self.volumes[column])[0],
                        persistence=True,
                        persistence_type="session",
                    ),
                ]
            ),
        )

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.uuid("layout"),
                "content": (
                    "Dashboard displaying in place volumetric results "
                    "from a sensitivity study."
                ),
            },
            {
                "id": self.uuid("graph-wrapper"),
                "content": (
                    "Chart showing results for the current selection. "
                    "Different charts and options can be selected from the menu above. "
                    "Different sensitivities can be highlighted by clicking in the tornado plot."
                ),
            },
            {
                "id": self.uuid("table"),
                "content": (
                    "The table shows statistics per sensitivity parameter. "
                    "Rows can be filtered by searching, and sorted by "
                    "clicking on a column header."
                ),
            },
            {
                "id": self.uuid("response"),
                "content": "Select the volumetric calculation to display.",
            },
            {
                "id": self.uuid("plot-type"),
                "content": (
                    "Controls the type of the visualized chart. "
                    "Per realization shows bars per realization, "
                    "while the boxplot shows the range per sensitivity."
                ),
            },
            *self.tornadoplot.tour_steps,
            {
                "id": self.uuid("ensemble"),
                "content": (
                    "If several ensembles are available, the active ensemble "
                    "can be selected here."
                ),
            },
            {
                "id": self.uuid("source"),
                "content": (
                    "If volumes have been calculated for different grids. "
                    "E.g. geogrid and eclipsegrid, the active grid can be selected here."
                ),
            },
            {
                "id": self.uuid("filters"),
                "content": (
                    "Filter on different combinations of e.g. zones, facies and regions "
                    "(The options will vary dependent on what was included "
                    "in the calculation.)"
                ),
            },
        ]

    @property
    def vol_columns(self) -> List[str]:
        """List of all columns in dataframe"""
        return list(self.volumes.columns)

    @property
    def selectors(self) -> List[str]:
        """List of available selector columns in dframe"""
        return [x for x in InplaceVolumesOneByOne.FILTERS if x in self.vol_columns]

    @property
    def responses(self) -> List[str]:
        """List of available volume responses in dframe"""
        return [
            x
            for x in self.vol_columns
            if x not in self.selectors
            and x not in InplaceVolumesOneByOne.ENSEMBLE_COLUMNS
        ]

    @property
    def plot_selector(self) -> html.Div:
        """Radiobuttons to select plot type"""
        return html.Div(
            children=[
                html.Span("Plot type:", style={"font-weight": "bold"}),
                dcc.Dropdown(
                    id=self.uuid("plot-type"),
                    options=[
                        {"label": i, "value": i}
                        for i in [
                            "Per realization",
                            "Per sensitivity name",
                            "Per sensitivity case",
                        ]
                    ],
                    value="Per realization",
                    clearable=False,
                    persistence=True,
                    persistence_type="session",
                ),
            ]
        )

    @property
    def response_selector(self) -> html.Div:
        """Dropdown to select volumetric response"""
        return html.Div(
            children=html.Label(
                children=[
                    html.Span("Volumetric calculation:", style={"font-weight": "bold"}),
                    dcc.Dropdown(
                        id=self.uuid("response"),
                        options=[
                            {"label": volume_description(i), "value": i}
                            for i in self.responses
                        ],
                        clearable=False,
                        value=self.initial_response
                        if self.initial_response in self.responses
                        else self.responses[0],
                        persistence=True,
                        persistence_type="session",
                    ),
                ]
            ),
        )

    @property
    def filter_selectors(self) -> List[html.Div]:
        """Dropdowns for dataframe columns that can be filtered on (Zone, Region, etc)"""
        return [
            html.Div(
                children=[
                    html.Details(
                        open=True,
                        children=[
                            html.Summary(selector.lower().capitalize()),
                            wcc.Select(
                                id=self.selectors_id[selector],
                                options=[
                                    {"label": i, "value": i}
                                    for i in list(self.volumes[selector].unique())
                                ],
                                value=list(self.volumes[selector].unique()),
                                multi=True,
                                size=min(20, len(self.volumes[selector].unique())),
                                persistence=True,
                                persistence_type="session",
                            ),
                        ],
                    )
                ]
            )
            for selector in self.selectors
        ]

    @property
    def layout(self) -> html.Div:
        """Main layout"""
        return html.Div(
            children=[
                wcc.FlexBox(
                    id=self.uuid("layout"),
                    children=[
                        html.Div(
                            children=[
                                wcc.FlexBox(
                                    children=[
                                        html.Div(
                                            style={"flex": 1},
                                            children=[
                                                self.selector(
                                                    "Ensemble", "ensemble", "ENSEMBLE"
                                                ),
                                                self.selector(
                                                    "Grid source", "source", "SOURCE"
                                                ),
                                                self.response_selector,
                                                self.plot_selector,
                                                html.Span(
                                                    "Filters:",
                                                    style={"font-weight": "bold"},
                                                ),
                                                html.Div(
                                                    id=self.uuid("filters"),
                                                    children=self.filter_selectors,
                                                ),
                                            ],
                                        ),
                                        html.Div(
                                            style={"flex": 3},
                                            children=[
                                                html.Div(
                                                    style={"height": "600px"},
                                                    id=self.uuid("graph-wrapper"),
                                                ),
                                                html.Div(
                                                    children=[
                                                        html.Div(
                                                            id=self.uuid(
                                                                "volume_title"
                                                            ),
                                                            style={
                                                                "textAlign": "center"
                                                            },
                                                            children="",
                                                        ),
                                                        DataTable(
                                                            id=self.uuid("table"),
                                                            sort_action="native",
                                                            filter_action="native",
                                                            page_action="native",
                                                            page_size=10,
                                                        ),
                                                    ],
                                                ),
                                            ],
                                        ),
                                    ],
                                )
                            ]
                        ),
                        html.Div(
                            id=self.uuid("tornado-wrapper"),
                            style={"visibility": "visible", "flex": 2},
                            children=[self.tornadoplot.layout],
                        ),
                    ],
                ),
            ]
        )

    def set_callbacks(self, app: dash.Dash) -> None:
        @app.callback(
            [
                Output(self.tornadoplot.storage_id, "data"),
                Output(self.uuid("table"), "data"),
                Output(self.uuid("table"), "columns"),
            ],
            [
                Input(i, "value")
                for sublist in [
                    [
                        self.uuid("ensemble"),
                        self.uuid("response"),
                        self.uuid("source"),
                    ],
                    list(self.selectors_id.values()),
                ]
                for i in sublist
            ],
        )
        def _render_table_and_tornado(
            ensemble: str, response: str, source: str, *filters: Union[str, List[str]]
        ) -> Tuple[str, List[dict], List[dict]]:
            # Filter data
            data = filter_dataframe(
                self.volumes, self.selectors, ensemble, source, filters
            )

            # Table data
            table, columns = calculate_table(data, response)

            # TornadoPlot input
            tornado = json.dumps(
                {
                    "ENSEMBLE": ensemble,
                    "data": data.groupby("REAL")
                    .sum()
                    .reset_index()[["REAL", response]]
                    .values.tolist(),
                    "number_format": "#.4g",
                    "unit": volume_unit(response),
                }
            )
            return tornado, table, columns

        @app.callback(
            [
                Output(self.uuid("graph-wrapper"), "children"),
                Output(self.uuid("volume_title"), "children"),
            ],
            [
                Input(self.tornadoplot.click_id, "data"),
                Input(self.tornadoplot.high_low_storage_id, "data"),
                Input(self.uuid("plot-type"), "value"),
            ],
            [
                State(i, "value")
                for sublist in [
                    [
                        self.uuid("ensemble"),
                        self.uuid("response"),
                        self.uuid("source"),
                    ],
                    list(self.selectors_id.values()),
                ]
                for i in sublist
            ],
        )
        def _render_chart(
            tornado_click_data_str: Union[str, None],
            high_low_storage: dict,
            plot_type: str,
            ensemble: str,
            response: str,
            source: str,
            *filters: Union[str, List[str]],
        ) -> Tuple[Union[wcc.Graph, html.Div], str]:
            if dash.callback_context.triggered is None:
                raise PreventUpdate

            tornado_click: Union[dict, None] = (
                json.loads(tornado_click_data_str) if tornado_click_data_str else None
            )

            ctx = dash.callback_context.triggered[0]["prop_id"].split(".")[0]
            if tornado_click:
                if (
                    high_low_storage is not None
                    and ctx == self.tornadoplot.high_low_storage_id
                ):
                    # Tornado plot is updated without a click in the plot, updating reals:
                    if tornado_click["sens_name"] in high_low_storage:
                        tornado_click["real_low"] = high_low_storage[
                            tornado_click["sens_name"]
                        ].get("real_low")
                        tornado_click["real_high"] = high_low_storage[
                            tornado_click["sens_name"]
                        ].get("real_high")
                    else:
                        # fallback in a case where chosen sens_name does not exist in updated
                        # tornado plot.
                        # (can this ever occur except when sens_name already is None?)
                        tornado_click["sens_name"] = None
                        tornado_click["real_low"] = []
                        tornado_click["real_high"] = []

            # Filter data
            data = filter_dataframe(
                self.volumes, self.selectors, ensemble, source, filters
            )

            # Volume title:
            volume_title = f"{volume_description(response)} [{volume_unit(response)}]"

            # Make Plotly figure
            layout = {}
            layout.update({"height": 600, "margin": {"l": 100, "b": 100}})
            if plot_type == "Per realization":
                # One bar per realization
                layout.update(
                    {
                        "xaxis": {"title": "Realizations"},
                        "yaxis": {"title": volume_title},
                    }
                )
                plot_data = data.groupby("REAL").sum().reset_index()

                if tornado_click:
                    figure_data = [
                        {
                            "y": plot_data[
                                plot_data["REAL"].isin(tornado_click["real_low"])
                            ][response],
                            "x": plot_data[
                                plot_data["REAL"].isin(tornado_click["real_low"])
                            ]["REAL"],
                            "tickformat": ".4s",
                            "type": "bar",
                            "showlegend": False,
                            "name": "",
                        },
                        {
                            "y": plot_data[
                                plot_data["REAL"].isin(tornado_click["real_high"])
                            ][response],
                            "x": plot_data[
                                plot_data["REAL"].isin(tornado_click["real_high"])
                            ]["REAL"],
                            "tickformat": ".4s",
                            "type": "bar",
                            "showlegend": False,
                            "name": "",
                        },
                        {
                            "y": plot_data[
                                ~plot_data["REAL"].isin(
                                    tornado_click["real_low"]
                                    + tornado_click["real_high"]
                                )
                            ][response],
                            "x": plot_data[
                                ~plot_data["REAL"].isin(
                                    tornado_click["real_low"]
                                    + tornado_click["real_high"]
                                )
                            ]["REAL"],
                            "tickformat": ".4s",
                            "type": "bar",
                            "marker": {"color": "grey"},
                            "showlegend": False,
                            "name": "",
                        },
                    ]
                else:
                    figure_data = [
                        {
                            "y": plot_data[response],
                            "x": plot_data["REAL"],
                            "tickformat": ".4s",
                            "type": "bar",
                            "marker": {"color": "grey"},
                        },
                    ]
                figure = wcc.Graph(
                    config={"displayModeBar": False},
                    id=self.uuid("graph"),
                    figure={
                        "data": figure_data,
                        "layout": self.theme.create_themed_layout(layout),
                    },
                )
            elif plot_type == "Per sensitivity case":
                # One box per sensitivity name
                layout.update(
                    {
                        "title": "Distribution for each sensitivity case",
                        "yaxis": {"title": volume_title},
                    }
                )
                figure = wcc.Graph(
                    config={"displayModeBar": False},
                    id=self.uuid("graph"),
                    figure={
                        "data": [
                            {
                                "y": senscase_df.groupby("REAL")
                                .sum()
                                .reset_index()[response],
                                "name": f"{sensname} ({senscase})",
                                "type": "box",
                            }
                            for sensname, sensname_df in data.groupby(["SENSNAME"])
                            for senscase, senscase_df in sensname_df.groupby(
                                ["SENSCASE"]
                            )
                        ],
                        "layout": self.theme.create_themed_layout(layout),
                    },
                )
            elif plot_type == "Per sensitivity name":
                # One box per sensitivity name
                layout.update(
                    {
                        "title": "Distribution for each sensitivity name",
                        "yaxis": {"title": volume_title},
                    }
                )
                figure = wcc.Graph(
                    config={"displayModeBar": False},
                    id=self.uuid("graph"),
                    figure={
                        "data": [
                            {
                                "y": sensname_df.groupby("REAL")
                                .sum()
                                .reset_index()[response],
                                "name": f"{sensname}",
                                "type": "box",
                            }
                            for sensname, sensname_df in data.groupby(["SENSNAME"])
                        ],
                        "layout": self.theme.create_themed_layout(layout),
                    },
                )
            else:
                # Should not occur unless plot_type options are changed
                figure = html.Div("Invalid plot type")

            return figure, volume_title


def calculate_table(df: pd.DataFrame, response: str) -> Tuple[List[dict], List[dict]]:
    table: List[dict] = []
    for (sensname, senscase), dframe in df.groupby(["SENSNAME", "SENSCASE"]):
        values = dframe.groupby("REAL").sum().reset_index()[response]
        try:
            table.append(
                {
                    "Sens Name": str(sensname),
                    "Sens Case": str(senscase),
                    "Minimum": values.min(),
                    "Maximum": values.max(),
                    "Mean": values.mean(),
                    "Stddev": values.std(),
                    "P10": np.percentile(values, 90),
                    "P90": np.percentile(values, 10),
                }
            )
        except KeyError:
            pass
    columns: List[dict] = [
        {**{"name": i[0], "id": i[0]}, **i[1]}
        for i in InplaceVolumesOneByOne.TABLE_STATISTICS
    ]
    return table, columns


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def filter_dataframe(
    dframe: pd.DataFrame,
    columns: Union[str, List[str]],
    ensemble: str,
    source: str,
    column_values: Iterable[Union[str, List[str]]],
) -> pd.DataFrame:
    df = dframe.copy()
    if not isinstance(columns, list):
        columns = [columns]
    for filt, col in zip(column_values, columns):
        if isinstance(filt, list):
            df = df.loc[df[col].isin(filt)]
        else:
            df = df.loc[df[col] == filt]
    df = df.loc[df["ENSEMBLE"] == ensemble]
    df = df.loc[df["SOURCE"] == source]
    return df


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file: Path) -> pd.DataFrame:
    return pd.read_csv(csv_file, index_col=None)



from typing import Optional, Union, List, Dict, Any
import io
import base64

import numpy as np
from xtgeo import RegularSurface
from webviz_config.common_cache import CACHE
from PIL import Image
from .image_processing import array_to_png


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def load_surface(surface_path: str) -> RegularSurface:
    return RegularSurface(surface_path)


def get_surface_arr(
    surface: RegularSurface,
    unrotate: bool = True,
    flip: bool = True,
    clip_min: Union[float, np.ndarray, None] = None,
    clip_max: Union[float, np.ndarray, None] = None,
) -> List[np.ndarray]:
    if clip_min or clip_max:
        np.ma.clip(surface.values, clip_min, clip_max, out=surface.values)
    if unrotate:
        surface.unrotate()

    x, y, z = surface.get_xyz_values()
    if flip:
        x = np.flip(x.transpose(), axis=0)
        y = np.flip(y.transpose(), axis=0)
        z = np.flip(z.transpose(), axis=0)
    z.filled(np.nan)
    return [x, y, z]


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_surface_fence(fence: np.ndarray, surface: RegularSurface) -> np.ndarray:
    return surface.get_fence(fence)


# pylint: disable=too-many-arguments
def make_surface_layer(
    surface: RegularSurface,
    name: str = "surface",
    updatemode: str = "update",
    min_val: Optional[float] = None,
    max_val: Optional[float] = None,
    color: Optional[List[str]] = None,
    shader_type: Optional[str] = "soft-hillshading",
    shadows: bool = False,
    unit: str = " ",
) -> Dict[str, Any]:
    """Make NewLayeredMap surface image base layer
    Args:
        surface: an xtgeo surface object
        name: name of the surface object
        min_val: minimum value to be plotted in map
        max_val: maximum value to be plotted in map
        color: an array with colors as strings
        shader_type: determines shader in map
        unit: determines unit on the map axes
    Returns:
        A surface layer that can be plotted in NewLayeredMap
    """

    zvalues = get_surface_arr(surface, clip_min=min_val, clip_max=max_val)[2]
    min_val = min_val if min_val is not None else np.nanmin(zvalues)
    max_val = max_val if max_val is not None else np.nanmax(zvalues)
    if shader_type == "hillshading_shadows":
        shader_type = "hillshading"
        shadows = True
    img = Image.open(io.BytesIO(base64.b64decode(array_to_png(zvalues.copy())[22:])))
    width, height = img.size
    if width * height >= 300 * 300:
        scale = 1.0
    else:
        ratio = (1000 ** 2) / (width * height)
        scale = np.sqrt(ratio).round(2)
    color = (
        [
            "#fde725",
            "#b5de2b",
            "#6ece58",
            "#35b779",
            "#1f9e89",
            "#26828e",
            "#31688e",
            "#3e4989",
            "#482878",
            "#440154",
        ]
        if color is None
        else color
    )
    return {
        "name": name,
        "checked": True,
        "id": name,
        "action": updatemode,
        "baseLayer": True,
        "data": [
            {
                "type": "image",
                "url": array_to_png(zvalues.copy()),
                "colorScale": {
                    "colors": color,
                    "prefixZeroAlpha": False,
                    "scaleType": "linear",
                },
                "shader": {
                    "type": shader_type,
                    "shadows": shadows,
                    "shadowIterations": 128,
                    "elevationScale": -1.0,
                    "pixelScale": 11000,
                    "setBlackToAlpha": True,
                },
                "bounds": [[surface.xmin, surface.ymin], [surface.xmax, surface.ymax]],
                "minvalue": round(np.nanmin(zvalues), 4),
                "maxvalue": round(np.nanmax(zvalues), 4),
                "unit": str(unit),
                "imageScale": scale,
            }
        ],
    }


def get_surface_layers(
    switch: Dict[str, bool],
    surface_name: str,
    surfaces: List[RegularSurface],
    min_val: Optional[float] = None,
    max_val: Optional[float] = None,
) -> List[Dict[str, Any]]:
    """Creates layers in map view from all surfaces
    Args:
        switch: Toggle hillshading on/off
        surface_name: Name of surface
        surfaces: List containing a single surface with corresponding depth error, depth trend etc.
        min_val: Minimum z-value of surface
        max_val: Maximum z-value of surface
    Returns:
        layers: List of all surface layers
    """
    shader_type = "hillshading" if switch["value"] is True else None
    depth_list = [
        "Depth",
        "Depth uncertainty",
        "Depth residual",
        "Depth residual uncertainty",
        "Depth trend",
        "Depth trend uncertainty",
    ]
    layers = []
    for i, surface in enumerate(surfaces):
        if surface is not None:
            s_layer = make_surface_layer(
                surface,
                name=depth_list[i],
                min_val=min_val,
                max_val=max_val,
                color=[
                    "#440154",
                    "#482878",
                    "#3e4989",
                    "#31688e",
                    "#26828e",
                    "#1f9e89",
                    "#35b779",
                    "#6ece58",
                    "#b5de2b",
                    "#fde725",
                ],
                shader_type=shader_type,
            )
            s_layer["id"] = surface_name + " " + depth_list[i] + "-id"
            s_layer["action"] = "add"
            layers.append(s_layer)
    return layers



import io
import base64

import numpy as np
from PIL import Image


def array_to_png(tensor: np.ndarray, shift: bool = True, colormap: bool = False) -> str:
    # pylint: disable=too-many-branches
    """The layered map dash component takes in pictures as base64 data
    (or as a link to an existing hosted image). I.e. for containers wanting
    to create pictures on-the-fly from numpy arrays, they have to be converted
    to base64. This is an example function of how that can be done.

    1) Scale the input array (tensor) to the range 0-255.
    2) If shift=True and colormap=False, the 0 value in the scaled range
       is reserved for np.nan (while the actual data points utilize the
       range 1-255.

       If shift=True and colormap=True, the 0 value in the colormap range
       has alpha value equal to 0.0 (i.e. full transparency). This makes it
       possible for np.nan values in the actual map becoming transparent in
       the image.
    3) If the array is two-dimensional, the picture is stored as greyscale.
       Otherwise it is either stored as RGB or RGBA (depending on if the size
       of the third dimension is three or four, respectively).
    """

    tensor -= np.nanmin(tensor)

    if np.nanmax(tensor) != 0:
        if shift:
            tensor *= 254.0 / np.nanmax(tensor)
            tensor += 1.0
        else:
            tensor *= 255.0 / np.nanmax(tensor)

    tensor[np.isnan(tensor)] = 0

    if colormap:
        if tensor.shape[0] != 1:
            raise ValueError("The first dimension of a colormap tensor should be 1")
        if tensor.shape[1] != 256:
            raise ValueError("The second dimension of a colormap tensor should be 256")
        if tensor.shape[2] not in [3, 4]:
            raise ValueError(
                "The third dimension of a colormap tensor should be either 3 or 4"
            )
        if shift:
            if tensor.shape[2] != 4:
                raise ValueError(
                    "Can not shift a colormap which is not utilizing alpha channel"
                )
            tensor[0][0][3] = 0.0  # Make first color channel transparent

    if tensor.ndim == 2:
        image = Image.fromarray(np.uint8(tensor), "L")
    elif tensor.ndim == 3:
        if tensor.shape[2] == 3:
            image = Image.fromarray(np.uint8(tensor), "RGB")
        elif tensor.shape[2] == 4:
            image = Image.fromarray(np.uint8(tensor), "RGBA")
        else:
            raise ValueError(
                "Third dimension of tensor must have length 3 (RGB) or 4 (RGBA)"
            )
    else:
        raise ValueError("Incorrect number of dimensions in tensor")
    byte_io = io.BytesIO()
    image.save(byte_io, format="png")

    byte_io.seek(0)

    base64_data = base64.b64encode(byte_io.read()).decode("ascii")

    return f"data:image/png;base64,{base64_data}"



from pathlib import Path

import pytest
import numpy as np

from webviz_subsurface._models.ensemble_set_model import EnsembleSetModel


@pytest.mark.usefixtures("app")
def test_single_ensemble(testdata_folder):

    emodel = EnsembleSetModel(
        ensemble_paths={
            "iter-0": str(
                Path(
                    testdata_folder / "reek_history_match" / "realization-*" / "iter-0"
                )
            )
        }
    )
    assert emodel.ens_folders == {
        "iter-0": Path("webviz-subsurface-testdata") / "reek_history_match"
    }
    assert len(emodel.ensembles) == 1
    smry = emodel.load_smry()
    assert len(smry.columns) == 476
    assert len(smry["DATE"].unique()) == 480
    assert smry["ENSEMBLE"].unique() == ["iter-0"]
    assert smry["ENSEMBLE"].dtype == np.dtype("O")
    assert all(
        np.issubdtype(dtype, np.number)
        for dtype in smry.drop(["REAL", "ENSEMBLE", "DATE"], axis=1).dtypes
    )

    parameters = emodel.load_parameters()
    assert all(col in parameters.columns for col in ["ENSEMBLE", "REAL"])
    assert parameters["ENSEMBLE"].dtype == np.dtype("O")
    assert parameters["REAL"].dtype == np.dtype("int64")
    assert len(parameters.columns) == 28


@pytest.mark.usefixtures("app")
def test_smry_load_multiple_ensembles(testdata_folder):

    emodel = EnsembleSetModel(
        ensemble_paths={
            "iter-0": str(
                Path(
                    testdata_folder / "reek_history_match" / "realization-*" / "iter-0"
                )
            ),
            "iter-1": str(
                Path(
                    testdata_folder / "reek_history_match" / "realization-*" / "iter-1"
                )
            ),
            "iter-2": str(
                Path(
                    testdata_folder / "reek_history_match" / "realization-*" / "iter-2"
                )
            ),
            "iter-3": str(
                Path(
                    testdata_folder / "reek_history_match" / "realization-*" / "iter-3"
                )
            ),
        }
    )
    smry = emodel.load_smry()
    assert len(smry.columns) == 476
    assert len(smry["DATE"].unique()) == 1141
    assert set(smry["ENSEMBLE"].unique()) == set(
        ["iter-0", "iter-1", "iter-2", "iter-3"]
    )
    assert smry["ENSEMBLE"].dtype == np.dtype("O")
    # assert smry["DATE"].dtype == np.dtype("O") # Fails due to wrong input data?
    assert smry["REAL"].dtype == np.dtype("int64")
    assert all(
        np.issubdtype(dtype, np.number)
        for dtype in smry.drop(["REAL", "ENSEMBLE", "DATE"], axis=1).dtypes
    )
    smeta = emodel.load_smry_meta()
    assert set(smeta.columns) == set(
        ["unit", "is_total", "is_rate", "is_historical", "keyword", "wgname", "get_num"]
    )
    assert len(smeta) == 473
    assert "FOPT" in smeta.index

    parameters = emodel.load_parameters()
    assert all(col in parameters.columns for col in ["ENSEMBLE", "REAL"])
    assert parameters["ENSEMBLE"].dtype == np.dtype("O")
    assert parameters["REAL"].dtype == np.dtype("int64")
    assert len(parameters.columns) == 28

    dframe = emodel.load_csv(Path("share") / "results" / "tables" / "rft.csv")
    assert "ENSEMBLE" in dframe.columns
    assert dframe["ENSEMBLE"].dtype == np.dtype("O")
    assert len(dframe["ENSEMBLE"].unique()) == 4
    assert len(dframe.columns) == 15

    with pytest.raises(KeyError) as exc:
        emodel.load_csv("some_path")
    assert (
        exc.value.args[0]
        == "No data found for load_csv with arguments: {'csv_file': 'some_path'}"
    )


@pytest.mark.usefixtures("app")
def test_webvizstore(testdata_folder):
    emodel = EnsembleSetModel(
        ensemble_paths={
            "iter-0": str(
                Path(
                    testdata_folder / "reek_history_match" / "realization-*" / "iter-0"
                )
            ),
            "iter-1": str(
                Path(
                    testdata_folder / "reek_history_match" / "realization-*" / "iter-1"
                )
            ),
            "iter-2": str(
                Path(
                    testdata_folder / "reek_history_match" / "realization-*" / "iter-2"
                )
            ),
            "iter-3": str(
                Path(
                    testdata_folder / "reek_history_match" / "realization-*" / "iter-3"
                )
            ),
        }
    )
    emodel.load_parameters()
    assert len(emodel.webvizstore) == 4
    emodel.load_smry()
    assert len(emodel.webvizstore) == 8
    emodel.load_smry_meta()
    assert len(emodel.webvizstore) == 12
    emodel.load_csv(Path("share") / "results" / "tables" / "rft.csv")
    assert len(emodel.webvizstore) == 16



def hex_to_rgba(hex_string: str, opacity: float = 1.0) -> str:
    """Converts the given hex color to rgba"""
    hex_string = hex_string.lstrip("#")
    hlen = len(hex_string)
    rgb = [int(hex_string[i : i + hlen // 3], 16) for i in range(0, hlen, hlen // 3)]
    return f"rgba({rgb[0]}, {rgb[1]}, {rgb[2]}, {opacity})"



from typing import Optional, List, Dict, Tuple, Callable, Any, Union
from pathlib import Path

import dash
import numpy as np
import pandas as pd
import dash_html_components as html
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State
import dash_core_components as dcc
import webviz_core_components as wcc
from webviz_config import WebvizPluginABC
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore

from webviz_subsurface._utils.unique_theming import unique_colors
from webviz_subsurface._datainput.fmu_input import load_csv
from ._processing import filter_frame
from ._formation_figure import FormationFigure
from ._map_figure import MapFigure
from ._misfit_figure import update_misfit_plot
from ._crossplot_figure import update_crossplot
from ._errorplot_figure import update_errorplot


class RftPlotter(WebvizPluginABC):
    """This plugin visualizes simulated RFT results from
FMU ensembles combined with ERT observation data.

Several visualizations are available:

* Map view of RFT observations.

* Depth vs pressure plot showing simulated RFT data along wells together with observation points.

* Barchart showing sum of mean misfit for ERT observations per realization. One plot per ensemble.

* Crossplot of simulated RFT vs observed value per ERT observation. One plot per ensemble.

* Boxplot showing misfit per ERT observation for each ensemble.

---
**Using data per realization**

* **`ensembles`**: Which ensembles in `shared_settings` to visualize.

In addition, you need to have rft-files in your realizations stored at the local path \
`share/results/tables`. The `rft_ert.csv` is required as input, while the `rft.csv` is optional:

* **`rft_ert.csv`**: A csv file containing simulated and observed RFT data for RFT observations \
defined in ERT \
[(example file)](https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
reek_history_match/realization-0/iter-0/share/results/tables/rft_ert.csv).

* **`rft.csv`**: A csv file containing simulated RFT data extracted from ECLIPSE RFT output files \
using [ecl2df](https://equinor.github.io/ecl2df/ecl2df.html#module-ecl2df.rft) \
[(example file)](https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
reek_history_match/realization-0/iter-0/share/results/tables/rft.csv). \
Simulated RFT data can be visualized along MD if a "CONMD" column is present in \
the dataframe and only for wells where each RFT datapoint has a unique MD.

**Using aggregated data**

* **`csvfile_rft`**: Aggregated version of `rft.csv` [(example file)](\
https://github.com/equinor/webviz-subsurface-testdata/blob/master/aggregated_data/rft.csv).
* **`csvfile_rft_ert`**: Aggregated version of `rft_ert.csv` [(example file)](\
https://github.com/equinor/webviz-subsurface-testdata/blob/master/aggregated_data/rft_ert.csv).


**Optional input for both input options**

* **`obsdata`**: A csv file containing additional RFT observation data not defined in ERT for
visualization together with simulated RFT.
Mandatory column names: `WELL`, `DATE` (yyyy-mm-dd), `DEPTH` and `PRESSURE`

* **`formations`**: A csv file containing top and base values for each zone per well.
Used to visualize zone boundaries together with simulated RFT.
Mandatory column names: `WELL`, `ZONE`, `TOP_TVD`, `BASE_TVD` \
[(example file))](\
https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
reek_history_match/share/results/tables/formations.csv).

* **`faultlines`**: A csv file containing faultpolygons to be visualized together with the map view.
Export format from [xtgeo.xyz.polygons.dataframe](
https://xtgeo.readthedocs.io/en/latest/apiref/xtgeo.xyz.polygons.html#xtgeo.xyz.polygons.Polygons.dataframe
) \
[(example file)](\
https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
reek_history_match/share/results/polygons/faultpolygons.csv).

---
?> Well name needs to be consistent with Eclipse well name.

?> Only RFT observations marked as active in ERT are used to generate plots.

The `rft_ert.csv` file can be generated by running the "MERGE_RFT_ERTOBS" forward model in ERT, \
this will collect ERT RFT observations and merge with CSV output from the "GENDATA_RFT" forward \
model. [ERT docs](https://fmu-docs.equinor.com/docs/ert/reference/\
forward_models.html?highlight=gendata_rft#MERGE_RFT_ERTOBS).

"""

    def __init__(
        self,
        app: dash.Dash,
        csvfile_rft: Path = None,
        csvfile_rft_ert: Path = None,
        ensembles: Optional[List[str]] = None,
        formations: Path = None,
        obsdata: Path = None,
        faultlines: Path = None,
    ) -> None:
        super().__init__()
        self.formations = formations
        self.faultlines = faultlines
        self.obsdata = obsdata
        self.csvfile_rft = csvfile_rft
        self.csvfile_rft_ert = csvfile_rft_ert

        self.simdf = read_csv(self.csvfile_rft) if csvfile_rft is not None else None
        self.formationdf = read_csv(self.formations) if self.formations else None
        self.faultlinesdf = read_csv(self.faultlines) if self.faultlines else None
        self.obsdatadf = read_csv(self.obsdata) if self.obsdata else None

        if csvfile_rft_ert and ensembles:
            raise ValueError(
                'Incorrent arguments. Either provide a "csvfile_rft_ert" or "ensembles"'
            )

        if csvfile_rft_ert is not None:
            self.ertdatadf = read_csv(self.csvfile_rft_ert)

        if ensembles:
            self.ens_paths = (
                {
                    ens: app.webviz_settings["shared_settings"]["scratch_ensembles"][
                        ens
                    ]
                    for ens in ensembles
                }
                if ensembles is not None
                else None
            )

            try:
                self.simdf = load_csv(self.ens_paths, "share/results/tables/rft.csv")
            except (KeyError, OSError):
                self.simdf = None

            try:
                self.ertdatadf = load_csv(
                    self.ens_paths, "share/results/tables/rft_ert.csv"
                )
            except KeyError as exc:
                raise KeyError(
                    "CSV file for ERT RFT observations/simulations "
                    "(share/results/tables/rft_ert.csv) not found!"
                ) from exc

        self.ertdatadf = self.ertdatadf.rename(
            columns={
                "time": "DATE",
                "is_active": "ACTIVE",
                "isactive": "ACTIVE",
                "well": "WELL",
                "zone": "ZONE",
                "pressure": "SIMULATED",
                "true_vertical_depth": "TVD",
                "measured_depth": "MD",
                "observed": "OBSERVED",
                "obs": "OBSERVED",
                "error": "OBSERVED_ERR",
                "utm_x": "EAST",
                "utm_y": "NORTH",
            }
        )
        self.ertdatadf["DIFF"] = (
            self.ertdatadf["SIMULATED"] - self.ertdatadf["OBSERVED"]
        )
        self.ertdatadf["ABSDIFF"] = abs(
            self.ertdatadf["SIMULATED"] - self.ertdatadf["OBSERVED"]
        )
        self.ertdatadf["YEAR"] = pd.to_datetime(self.ertdatadf["DATE"]).dt.year
        self.ertdatadf = self.ertdatadf.sort_values(by="DATE")
        self.ertdatadf["DATE_IDX"] = self.ertdatadf["DATE"].apply(
            lambda x: list(self.ertdatadf["DATE"].unique()).index(x)
        )
        self.date_marks = self.set_date_marks()
        self.ertdatadf = filter_frame(
            self.ertdatadf,
            {
                "ACTIVE": 1,
            },
        )
        self.ertdatadf["STDDEV"] = self.ertdatadf.groupby(
            ["WELL", "DATE", "ZONE", "ENSEMBLE", "TVD"]
        )["SIMULATED"].transform("std")

        self.set_callbacks(app)

    def add_webvizstore(self) -> List[Tuple[Callable, List[Dict[str, Any]]]]:
        functions: List[Tuple[Callable, List[Dict[str, Any]]]] = [
            (
                read_csv,
                [
                    {"csv_file": path}
                    for path in [
                        self.faultlines,
                        self.formations,
                        self.obsdata,
                        self.csvfile_rft,
                        self.csvfile_rft_ert,
                    ]
                    if path is not None
                ],
            )
        ]
        if self.csvfile_rft_ert is None:
            functions.append(
                (
                    load_csv,
                    [
                        {
                            "ensemble_paths": self.ens_paths,
                            "csv_file": "share/results/tables/rft_ert.csv",
                        },
                    ],
                )
            )
            try:
                load_csv(self.ens_paths, "share/results/tables/rft.csv")
                functions.append(
                    (
                        load_csv,
                        [
                            {
                                "ensemble_paths": self.ens_paths,
                                "csv_file": "share/results/tables/rft.csv",
                            },
                        ],
                    )
                )
            except KeyError:
                pass

        return functions

    @property
    def well_names(self) -> List[str]:
        return sorted(list(self.ertdatadf["WELL"].unique()))

    @property
    def zone_names(self) -> List[str]:
        return sorted(list(self.ertdatadf["ZONE"].unique()))

    @property
    def dates(self) -> List[str]:
        return sorted(list(self.ertdatadf["DATE"].unique()))

    def date_in_well(self, well: str) -> List[str]:
        df = self.ertdatadf.loc[self.ertdatadf["WELL"] == well]
        return [str(d) for d in list(df["DATE"].unique())]

    @property
    def ensembles(self) -> List[str]:
        return list(self.ertdatadf["ENSEMBLE"].unique())

    @property
    def enscolors(self) -> dict:
        return unique_colors(self.ensembles)

    def set_date_marks(self) -> Dict[str, Dict[str, Any]]:
        marks = {}
        idx_steps = np.linspace(
            start=0,
            stop=self.ertdatadf["DATE_IDX"].max(),
            num=min(5, len(self.ertdatadf["DATE_IDX"].unique())),
            dtype=int,
        )
        date_steps = self.ertdatadf.loc[self.ertdatadf["DATE_IDX"].isin(idx_steps)][
            "DATE"
        ].unique()

        for i, date_index in enumerate(idx_steps):
            marks[str(date_index)] = {
                "label": f"{date_steps[i]}",
                "style": {
                    "white-space": "nowrap",
                    "font-weight": "bold",
                },
            }
        return marks

    @property
    def formation_plot_selectors(self) -> List[html.Div]:
        return [
            html.Div(
                [
                    html.Label(
                        style={"font-weight": "bold"},
                        children="Ensembles in well plot",
                    ),
                    dcc.Dropdown(
                        id=self.uuid("ensemble"),
                        options=[
                            {"label": ens, "value": ens} for ens in self.ensembles
                        ],
                        value=self.ensembles[0],
                        multi=True,
                        clearable=False,
                        persistence=True,
                        persistence_type="session",
                    ),
                ]
            ),
            wcc.FlexBox(
                children=[
                    html.Div(
                        style={"flex": 1},
                        children=[
                            html.Label(
                                style={"font-weight": "bold"},
                                children="Well",
                            ),
                            dcc.Dropdown(
                                id=self.uuid("well"),
                                options=[
                                    {"label": well, "value": well}
                                    for well in self.well_names
                                ],
                                value=self.well_names[0],
                                clearable=False,
                                persistence=True,
                                persistence_type="session",
                            ),
                        ],
                    ),
                    html.Div(
                        style={"flex": 1},
                        children=[
                            html.Label(
                                style={"font-weight": "bold"},
                                children="Date",
                            ),
                            dcc.Dropdown(
                                id=self.uuid("date"),
                                options=[
                                    {"label": date, "value": date}
                                    for date in self.date_in_well(self.well_names[0])
                                ],
                                clearable=False,
                                value=self.date_in_well(self.well_names[0])[0],
                                persistence=True,
                                persistence_type="session",
                            ),
                        ],
                    ),
                ]
            ),
            wcc.FlexBox(
                children=[
                    html.Div(
                        style={"flex": 1},
                        children=[
                            html.Label(
                                style={"font-weight": "bold"},
                                children="Plot simulated results as",
                            ),
                            dcc.RadioItems(
                                id=self.uuid("linetype"),
                                options=[
                                    {
                                        "label": "Realization lines",
                                        "value": "realization",
                                    },
                                    {
                                        "label": "Statistical fanchart",
                                        "value": "fanchart",
                                    },
                                ],
                                value="realization",
                                labelStyle={
                                    "display": "inline-block",
                                    "margin": "5px",
                                },
                                persistence=True,
                                persistence_type="session",
                            ),
                        ],
                    ),
                    html.Div(
                        style={"flex": 1},
                        children=[
                            html.Label(
                                style={"font-weight": "bold"},
                                children="Depth option",
                            ),
                            dcc.RadioItems(
                                id=self.uuid("depth_option"),
                                options=[
                                    {
                                        "label": "TVD",
                                        "value": "TVD",
                                    },
                                    {
                                        "label": "MD",
                                        "value": "MD",
                                    },
                                ],
                                value="TVD",
                                labelStyle={
                                    "display": "inline-block",
                                    "margin": "5px",
                                },
                                persistence=True,
                                persistence_type="session",
                            ),
                        ],
                    ),
                ],
            ),
        ]

    @property
    def map_plot_selectors(self) -> List[html.Div]:

        return [
            html.Div(
                style={"marginRight": "10px"},
                children=[
                    html.Div(
                        style={"width": "50%"},
                        children=[
                            html.Label(
                                style={"font-weight": "bold"},
                                children="Ensemble in map plot",
                            ),
                            dcc.Dropdown(
                                id=self.uuid("map_ensemble"),
                                options=[
                                    {"label": ens, "value": ens}
                                    for ens in list(self.ertdatadf["ENSEMBLE"].unique())
                                ],
                                value=list(self.ertdatadf["ENSEMBLE"].unique())[0],
                                clearable=False,
                                persistence=True,
                                persistence_type="session",
                            ),
                        ],
                    ),
                    wcc.FlexBox(
                        children=[
                            html.Div(
                                style={"flex": 1},
                                children=[
                                    html.Label(
                                        style={"font-weight": "bold"},
                                        children="Size by",
                                    ),
                                    dcc.Dropdown(
                                        id=self.uuid("map_size"),
                                        options=[
                                            {
                                                "label": "Standard Deviation",
                                                "value": "STDDEV",
                                            },
                                            {
                                                "label": "Misfit",
                                                "value": "ABSDIFF",
                                            },
                                        ],
                                        value="ABSDIFF",
                                        clearable=False,
                                        persistence=True,
                                        persistence_type="session",
                                    ),
                                ],
                            ),
                            html.Div(
                                style={"flex": 1},
                                children=[
                                    html.Label(
                                        style={"font-weight": "bold"},
                                        children="Color by",
                                    ),
                                    dcc.Dropdown(
                                        id=self.uuid("map_color"),
                                        options=[
                                            {
                                                "label": "Misfit",
                                                "value": "ABSDIFF",
                                            },
                                            {
                                                "label": "Standard Deviation",
                                                "value": "STDDEV",
                                            },
                                            {
                                                "label": "Year",
                                                "value": "YEAR",
                                            },
                                        ],
                                        value="STDDEV",
                                        clearable=False,
                                        persistence=True,
                                        persistence_type="session",
                                    ),
                                ],
                            ),
                        ]
                    ),
                    html.Label(
                        style={"font-weight": "bold"},
                        children="Date range",
                    ),
                    html.Div(
                        style={"width": "100%", "height": "70px"},
                        children=[
                            dcc.RangeSlider(
                                id=self.uuid("map_date"),
                                min=self.ertdatadf["DATE_IDX"].min(),
                                max=self.ertdatadf["DATE_IDX"].max(),
                                value=[
                                    self.ertdatadf["DATE_IDX"].min(),
                                    self.ertdatadf["DATE_IDX"].max(),
                                ],
                                marks=self.date_marks,
                                persistence=True,
                                persistence_type="session",
                            )
                        ],
                    ),
                ],
            ),
        ]

    def filter_layout(
        self, tab: str
    ) -> List[dash.development.base_component.Component]:
        """Layout for shared filters"""
        return [
            html.Label(
                style={"font-weight": "bold"},
                children=["Ensembles"],
            ),
            wcc.Select(
                size=min(4, len(self.ensembles)),
                id=self.uuid(f"ensemble-{tab}"),
                options=[{"label": name, "value": name} for name in self.ensembles],
                value=self.ensembles,
                multi=True,
                persistence=True,
                persistence_type="session",
            ),
            html.Label(
                style={"font-weight": "bold"},
                children=["Wells"],
            ),
            wcc.Select(
                size=min(20, len(self.well_names)),
                id=self.uuid(f"well-{tab}"),
                options=[{"label": name, "value": name} for name in self.well_names],
                value=self.well_names,
                multi=True,
                persistence=True,
                persistence_type="session",
            ),
            html.Label(
                style={"font-weight": "bold"},
                children=["Zones"],
            ),
            wcc.Select(
                size=min(10, len(self.zone_names)),
                id=self.uuid(f"zone-{tab}"),
                options=[{"label": name, "value": name} for name in self.zone_names],
                value=self.zone_names,
                multi=True,
                persistence=True,
                persistence_type="session",
            ),
            html.Label(
                style={"font-weight": "bold"},
                children=["Dates"],
            ),
            wcc.Select(
                size=min(10, len(self.dates)),
                id=self.uuid(f"date-{tab}"),
                options=[{"label": name, "value": name} for name in self.dates],
                value=self.dates,
                multi=True,
                persistence=True,
                persistence_type="session",
            ),
        ]

    def size_color_layout(self) -> List[html.Div]:
        return [
            html.Div(
                children=[
                    html.Label(
                        style={"font-weight": "bold"},
                        children="Color by",
                    ),
                    dcc.Dropdown(
                        id=self.uuid("crossplot_color"),
                        options=[
                            {
                                "label": "Misfit",
                                "value": "ABSDIFF",
                            },
                            {
                                "label": "Standard Deviation",
                                "value": "STDDEV",
                            },
                        ],
                        value="STDDEV",
                        clearable=False,
                    ),
                ],
            ),
            html.Div(
                children=[
                    html.Label(
                        style={"font-weight": "bold"},
                        children="Size by",
                    ),
                    dcc.Dropdown(
                        id=self.uuid("crossplot_size"),
                        options=[
                            {
                                "label": "Standard Deviation",
                                "value": "STDDEV",
                            },
                            {
                                "label": "Misfit",
                                "value": "ABSDIFF",
                            },
                        ],
                        value="ABSDIFF",
                        clearable=False,
                        persistence=True,
                        persistence_type="session",
                    ),
                ],
            ),
        ]

    @property
    def layout(self) -> dcc.Tabs:

        tabs_styles = {"height": "44px", "width": "100%"}
        tab_style = {
            "borderBottom": "1px solid #d6d6d6",
            "padding": "6px",
            "fontWeight": "bold",
        }

        tab_selected_style = {
            "borderTop": "1px solid #d6d6d6",
            "borderBottom": "1px solid #d6d6d6",
            "backgroundColor": "#007079",
            "color": "white",
            "padding": "6px",
        }

        return dcc.Tabs(
            style=tabs_styles,
            children=[
                dcc.Tab(
                    label="RFT Map",
                    style=tab_style,
                    selected_style=tab_selected_style,
                    children=[
                        html.Div(
                            children=[
                                wcc.FlexBox(
                                    children=[
                                        html.Div(
                                            style={"flex": 1},
                                            children=self.map_plot_selectors,
                                        ),
                                        html.Div(
                                            style={"flex": 1},
                                            children=self.formation_plot_selectors,
                                        ),
                                    ]
                                ),
                                wcc.FlexBox(
                                    children=[
                                        wcc.Graph(
                                            style={"flex": 1},
                                            id=self.uuid("map"),
                                        ),
                                        wcc.Graph(
                                            style={"flex": 1},
                                            id=self.uuid("graph"),
                                            figure={
                                                "layout": {
                                                    "height": 800,
                                                    "margin": {"t": 50},
                                                    "xaxis": {"showgrid": False},
                                                    "yaxis": {"showgrid": False},
                                                }
                                            },
                                        ),
                                    ]
                                ),
                            ]
                        )
                    ],
                ),
                dcc.Tab(
                    label="RFT misfit per real",
                    style=tab_style,
                    selected_style=tab_selected_style,
                    children=[
                        wcc.FlexBox(
                            children=[
                                html.Div(
                                    style={"flex": 1},
                                    children=self.filter_layout("misfitplot"),
                                ),
                                html.Div(
                                    style={"flex": 4},
                                    id=self.uuid("misfit-graph-wrapper"),
                                ),
                            ]
                        )
                    ],
                ),
                dcc.Tab(
                    label="RFT crossplot - sim vs obs",
                    style=tab_style,
                    selected_style=tab_selected_style,
                    children=[
                        wcc.FlexBox(
                            children=[
                                html.Div(
                                    style={"flex": 1},
                                    children=self.filter_layout("crossplot")
                                    + self.size_color_layout(),
                                ),
                                html.Div(
                                    style={"flex": 4},
                                    children=[
                                        html.Div(
                                            id=self.uuid("crossplot-graph-wrapper")
                                        ),
                                    ],
                                ),
                            ],
                        ),
                    ],
                ),
                dcc.Tab(
                    label="RFT misfit per observation",
                    style=tab_style,
                    selected_style=tab_selected_style,
                    children=[
                        wcc.FlexBox(
                            children=[
                                html.Div(
                                    style={"flex": 1},
                                    children=self.filter_layout("errorplot"),
                                ),
                                html.Div(
                                    style={"flex": 4},
                                    children=wcc.Graph(id=self.uuid("errorplot-graph")),
                                ),
                            ],
                        ),
                    ],
                ),
            ],
        )

    def set_callbacks(self, app: dash.Dash) -> None:
        @app.callback(
            Output(self.uuid("well"), "value"),
            [
                Input(self.uuid("map"), "clickData"),
            ],
        )
        def _get_clicked_well(click_data: Dict[str, List[Dict[str, Any]]]) -> str:
            if not click_data:
                return self.well_names[0]
            for layer in click_data["points"]:
                try:
                    return layer["customdata"]
                except KeyError:
                    pass
            raise PreventUpdate

        @app.callback(
            Output(self.uuid("map"), "figure"),
            [
                Input(self.uuid("map_ensemble"), "value"),
                Input(self.uuid("map_size"), "value"),
                Input(self.uuid("map_color"), "value"),
                Input(self.uuid("map_date"), "value"),
            ],
        )
        def _update_map(
            ensemble: str, sizeby: str, colorby: str, dates: List[float]
        ) -> Dict[str, Any]:
            figure = MapFigure(self.ertdatadf, ensemble)
            if self.faultlinesdf is not None:
                figure.add_fault_lines(self.faultlinesdf)
            figure.add_misfit_plot(sizeby, colorby, dates)

            return {"data": figure.traces, "layout": figure.layout}

        @app.callback(
            Output(self.uuid("graph"), "figure"),
            [
                Input(self.uuid("well"), "value"),
                Input(self.uuid("date"), "value"),
                Input(self.uuid("ensemble"), "value"),
                Input(self.uuid("linetype"), "value"),
                Input(self.uuid("depth_option"), "value"),
            ],
        )
        def _update_formation_plot(
            well: str, date: str, ensembles: List[str], linetype: str, depth_option: str
        ) -> Dict[str, Any]:
            if date not in self.date_in_well(well):
                raise PreventUpdate

            figure = FormationFigure(
                well=well,
                ertdf=self.ertdatadf,
                enscolors=self.enscolors,
                depth_option=depth_option,
                date=date,
                ensembles=ensembles,
                simdf=self.simdf,
                obsdf=self.obsdatadf,
            )

            if self.formations is not None:
                figure.add_formation(self.formationdf)

            figure.add_simulated_lines(linetype)
            figure.add_additional_observations()
            figure.add_ert_observed()

            return {
                "data": figure.traces,
                "layout": figure.layout,
            }

        @app.callback(
            Output(self.uuid("linetype"), "options"),
            Output(self.uuid("linetype"), "value"),
            Input(self.uuid("depth_option"), "value"),
            State(self.uuid("linetype"), "value"),
            State(self.uuid("well"), "value"),
            State(self.uuid("date"), "value"),
        )
        def _update_linetype(
            depth_option: str,
            current_linetype: str,
            current_well: str,
            current_date: str,
        ) -> Tuple[List[Dict[str, str]], str]:
            if self.simdf is not None:
                df = filter_frame(
                    self.simdf,
                    {"WELL": current_well, "DATE": current_date},
                )
                if depth_option == "TVD" or (
                    depth_option == "MD"
                    and "CONMD" in self.simdf
                    and len(df["CONMD"].unique()) == len(df["DEPTH"].unique())
                ):

                    return [
                        {
                            "label": "Realization lines",
                            "value": "realization",
                        },
                        {
                            "label": "Statistical fanchart",
                            "value": "fanchart",
                        },
                    ], current_linetype

            return [
                {
                    "label": "Realization lines",
                    "value": "realization",
                },
            ], "realization"

        @app.callback(
            [Output(self.uuid("date"), "options"), Output(self.uuid("date"), "value")],
            [
                Input(self.uuid("well"), "value"),
            ],
            [State(self.uuid("date"), "value")],
        )
        def _update_date(
            well: str, current_date: str
        ) -> Tuple[List[Dict[str, str]], str]:
            dates = self.date_in_well(well)
            available_dates = [{"label": date, "value": date} for date in dates]
            date = current_date if current_date in dates else dates[0]
            return available_dates, date

        @app.callback(
            Output(self.uuid("misfit-graph-wrapper"), "children"),
            [
                Input(self.uuid("well-misfitplot"), "value"),
                Input(self.uuid("zone-misfitplot"), "value"),
                Input(self.uuid("date-misfitplot"), "value"),
                Input(self.uuid("ensemble-misfitplot"), "value"),
            ],
        )
        def _misfit_plot(
            wells: List[str], zones: List[str], dates: List[str], ensembles: List[str]
        ) -> List[wcc.Graph]:
            df = filter_frame(
                self.ertdatadf,
                {"WELL": wells, "ZONE": zones, "DATE": dates, "ENSEMBLE": ensembles},
            )
            return update_misfit_plot(df, self.enscolors)

        @app.callback(
            Output(self.uuid("crossplot-graph-wrapper"), "children"),
            [
                Input(self.uuid("well-crossplot"), "value"),
                Input(self.uuid("zone-crossplot"), "value"),
                Input(self.uuid("date-crossplot"), "value"),
                Input(self.uuid("ensemble-crossplot"), "value"),
                Input(self.uuid("crossplot_size"), "value"),
                Input(self.uuid("crossplot_color"), "value"),
            ],
        )
        def _crossplot(
            wells: List[str],
            zones: List[str],
            dates: List[str],
            ensembles: List[str],
            sizeby: str,
            colorby: str,
        ) -> List[wcc.Graph]:
            df = filter_frame(
                self.ertdatadf,
                {"WELL": wells, "ZONE": zones, "DATE": dates, "ENSEMBLE": ensembles},
            )
            return update_crossplot(df, sizeby, colorby)

        @app.callback(
            Output(self.uuid("errorplot-graph"), "figure"),
            [
                Input(self.uuid("well-errorplot"), "value"),
                Input(self.uuid("zone-errorplot"), "value"),
                Input(self.uuid("date-errorplot"), "value"),
                Input(self.uuid("ensemble-errorplot"), "value"),
            ],
        )
        def _errorplot(
            wells: List[str], zones: List[str], dates: List[str], ensembles: List[str]
        ) -> Dict[str, Union[list, dict]]:
            df = filter_frame(
                self.ertdatadf,
                {"WELL": wells, "ZONE": zones, "DATE": dates, "ENSEMBLE": ensembles},
            )
            return update_errorplot(df, self.enscolors)


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file: str) -> pd.DataFrame:
    return pd.read_csv(csv_file)



from typing import Union, Optional, List, Callable, Tuple, Dict, Any
import pathlib

import pandas as pd

from .ensemble_model import EnsembleModel


class EnsembleSetModel:
    """Class to load and manipulate ensemble sets from given paths to
    ensembles on disk"""

    def __init__(
        self,
        ensemble_paths: dict,
        ensemble_set_name: str = "EnsembleSet",
        filter_file: Union[str, None] = "OK",
    ) -> None:
        self.ensemble_paths = ensemble_paths
        self.ensemble_set_name = ensemble_set_name
        self.filter_file = filter_file
        self._webvizstore: List = []
        self.ensembles = [
            EnsembleModel(ens_name, ens_path, filter_file=self.filter_file)
            for ens_name, ens_path in self.ensemble_paths.items()
        ]

    def __repr__(self) -> str:
        return f"EnsembleSetModel: {self.ensemble_paths}"

    @property
    def ens_folders(self) -> dict:
        """Get root folders for ensemble set"""
        return {
            ens: pathlib.Path(ens_path.split("realization")[0])
            for ens, ens_path in self.ensemble_paths.items()
        }

    def _get_ensembles_data(self, func: str, **kwargs: Any) -> pd.DataFrame:
        """Runs the provided function for each ensemble and concats dataframes"""
        dfs = []
        for ensemble in self.ensembles:
            try:
                dframe = getattr(ensemble, func)(**kwargs)
                dframe.insert(0, "ENSEMBLE", ensemble.ensemble_name)
                dfs.append(dframe)
            except (KeyError, ValueError):
                # Happens if an ensemble is missing some data
                # Warning has already been issued at initialization
                pass
        if dfs:
            return pd.concat(dfs, sort=False)
        raise KeyError(f"No data found for {func} with arguments: {kwargs}")

    def load_parameters(self) -> pd.DataFrame:
        return self._get_ensembles_data("load_parameters")

    def load_smry(
        self,
        time_index: Optional[Union[list, str]] = None,
        column_keys: Optional[list] = None,
    ) -> pd.DataFrame:
        return self._get_ensembles_data(
            "load_smry", time_index=time_index, column_keys=column_keys
        )

    def load_smry_meta(
        self,
        column_keys: Optional[list] = None,
    ) -> pd.DataFrame:
        """Finds metadata for the summary vectors in the ensemble set.
        Note that we assume the same units for all ensembles.
        (meaning that we update/overwrite when checking the next ensemble)
        """

        smry_meta: dict = {}
        for ensemble in self.ensembles:
            smry_meta.update(
                ensemble.load_smry_meta(column_keys=column_keys).T.to_dict()
            )
        return pd.DataFrame(smry_meta).transpose()

    def load_csv(self, csv_file: pathlib.Path) -> pd.DataFrame:
        return self._get_ensembles_data("load_csv", csv_file=csv_file)

    @property
    def webvizstore(self) -> List[Tuple[Callable, List[Dict]]]:
        store_functions = []
        for ensemble in self.ensembles:
            store_functions.extend(ensemble.webviz_store)
        return store_functions



from typing import Optional, List, Tuple, Callable
import shutil
import warnings
import pathlib
import datetime

import pandas as pd
import dash
import dash_html_components as html
import webviz_core_components as wcc
from webviz_config.webviz_store import webvizstore
from webviz_config.common_cache import CACHE
from webviz_config import WebvizPluginABC


class DiskUsage(WebvizPluginABC):
    """Visualize disk usage in a FMU project. It adds a dashboard showing disk usage per user.

    ---

    * **`scratch_dir`:** Path to the scratch directory to show disk usage for.
    * **`date`:** Date as string of form YYYY-MM-DD to request an explisit date. Default is to
    to use the most recent file avaialable, limited to the last week.

    ---

    ?> The `scratch_dir` directory must have a hidden folder `.disk_usage` containing daily
    csv files called `disk_usage_user_YYYY-MM-DD.csv`, where YYYY-MM-DD is the date.
    The plugin will search backwards from the current date, and throw an error if no file was found
    from the last week.

    The csv file must have the columns `userid` and `usageKB` (where KB means
    [kibibytes](https://en.wikipedia.org/wiki/Kibibyte)). All other columns are ignored.
    """

    def __init__(
        self, app: dash.Dash, scratch_dir: pathlib.Path, date: Optional["str"] = None
    ):

        super().__init__()

        self.scratch_dir = scratch_dir
        self.date_input = date
        self.disk_usage = get_disk_usage(self.scratch_dir, self.date_input)
        self.date = str(self.disk_usage["date"].unique()[0])
        self.users = self.disk_usage["userid"]
        self.usage_gib = self.disk_usage["usageKB"] / (1024 ** 2)
        self.theme = app.webviz_settings["theme"]

    @property
    def layout(self) -> html.Div:
        return html.Div(
            [
                html.H5(
                    f"Disk usage on {self.scratch_dir} per user as of {self.date}",
                    style={"text-align": "center"},
                ),
                wcc.FlexBox(
                    children=[
                        wcc.Graph(
                            style={"flex": 1},
                            figure=self.pie_chart,
                            config={"displayModeBar": False},
                        ),
                        wcc.Graph(style={"flex": 2}, figure=self.bar_chart),
                    ]
                ),
            ]
        )

    @property
    def pie_chart(self) -> dict:
        return {
            "data": [
                {
                    "values": self.usage_gib,
                    "labels": self.users,
                    "pull": (self.users.values == "<b>Free space</b>") * 0.05,
                    "text": (self.usage_gib).map("{:.2f} GiB".format),
                    "textinfo": "label",
                    "textposition": "inside",
                    "hoverinfo": "label+text",
                    "type": "pie",
                }
            ],
            "layout": self.theme.create_themed_layout({}),
        }

    @property
    def bar_chart(self) -> dict:
        return {
            "data": [
                {
                    "y": self.usage_gib,
                    "x": self.users,
                    "text": (self.usage_gib).map("{:.2f} GiB".format),
                    "hoverinfo": "x+text",
                    "type": "bar",
                }
            ],
            "layout": self.theme.create_themed_layout(
                {"yaxis": {"title": "GiB (GibiBytes)"}}
            ),
        }

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return [
            (
                get_disk_usage,
                [{"scratch_dir": self.scratch_dir, "date": self.date_input}],
            )
        ]


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def get_disk_usage(scratch_dir: pathlib.Path, date: Optional[str]) -> pd.DataFrame:
    def _loop_dates(scratch_dir: pathlib.Path) -> pd.DataFrame:
        today = datetime.datetime.today()
        for i in range(7):
            date = today - datetime.timedelta(days=i)
            try:
                return pd.read_csv(
                    scratch_dir
                    / ".disk_usage"
                    / f"disk_usage_user_{date.strftime('%Y-%m-%d')}.csv"
                ).assign(date=date.strftime("%Y-%m-%d"))
            except FileNotFoundError:
                continue
        raise FileNotFoundError(
            f"No disk usage file found for last week in {scratch_dir}."
        )

    if date is None:
        df = _loop_dates(scratch_dir)
    else:
        try:
            df = pd.read_csv(
                scratch_dir / ".disk_usage" / f"disk_usage_user_{date}.csv"
            ).assign(date=date)
        except FileNotFoundError as exc:
            raise FileNotFoundError(
                f"No disk usage file found for {date} in {scratch_dir}."
            ) from exc

    free_space_kib = (shutil.disk_usage(scratch_dir).total / 1024) - df["usageKB"].sum()

    if free_space_kib < 0:
        warnings.warn(
            f"Reported disk usage in "
            f"{scratch_dir}/.disk_usage/disk_usage_user_{df['date'].unique()[0]}.csv "
            f"must be wrong (unless total disk size has changed after {df['date'].unique()[0]}). "
            f"Total reported usage is {int(df['usageKB'].sum() / 1024**2)} GiB, "
            f"but the disk size is only {int(shutil.disk_usage(scratch_dir).total / 1024**3)} GiB "
            f"(this would imply negative free space of {int(free_space_kib / 1024**2)} GiB).",
            UserWarning,
            stacklevel=0,
        )
    else:
        df = df.append(
            {
                "userid": "<b>Free space</b>",
                "usageKB": free_space_kib,
                "date": df["date"].unique()[0],
            },
            ignore_index=True,
        )

    return df.sort_values(by="usageKB", axis=0, ascending=False)
